"""
Adapters to convert export DTOs to repository query objects.
Handles conversion of report-specific filter and order parameters.
"""

from __future__ import annotations

from collections.abc import Callable
from typing import TYPE_CHECKING, Any

import sqlalchemy as sa

from ai.backend.common.dto.manager.export import (
    AuditLogExportFilter,
    AuditLogExportOrder,
    BooleanFilter,
    DateTimeRangeFilter,
    OrderDirection,
    ProjectExportFilter,
    ProjectExportOrder,
    SessionExportFilter,
    SessionExportOrder,
    UserExportFilter,
    UserExportOrder,
)
from ai.backend.common.dto.manager.query import StringFilter
from ai.backend.manager.api.adapter import BaseFilterAdapter
from ai.backend.manager.api.gql.base import StringMatchSpec
from ai.backend.manager.errors.export import InvalidExportFieldKeys
from ai.backend.manager.repositories.base import QueryCondition, QueryOrder
from ai.backend.manager.repositories.base.export import (
    ExportFieldDef,
    ExportFieldType,
    JoinDef,
    ReportDef,
    StreamingExportQuery,
)

if TYPE_CHECKING:
    from sqlalchemy.orm.attributes import InstrumentedAttribute

__all__ = ("ExportAdapter",)


class ExportAdapter(BaseFilterAdapter):
    """Adapter for converting export request DTOs to query objects.

    Provides report-specific methods to build queries from filter and order DTOs.
    """

    def build_user_query(
        self,
        report: ReportDef,
        fields: list[str] | None,
        filter: UserExportFilter | None,
        order: list[UserExportOrder] | None,
        max_rows: int,
        statement_timeout_sec: int,
    ) -> StreamingExportQuery:
        """Build StreamingExportQuery for user export.

        Dynamically applies JOINs based on selected fields.
        Fields with joins attribute will trigger LEFT JOIN operations.

        Args:
            report: User report definition
            fields: Field keys to include (None = all fields)
            filter: User-specific filter conditions
            order: User-specific order specifications
            max_rows: Maximum rows to export
            statement_timeout_sec: Query timeout in seconds

        Returns:
            StreamingExportQuery ready for execution
        """
        selected_fields = self._select_fields(report, fields)
        conditions = self._build_user_conditions(report, filter)
        orders = self._build_user_orders(report, order)

        # Collect all required JOINs from selected fields
        all_joins = self._collect_joins(selected_fields)

        # Build select_from with dynamic JOINs
        select_from = self._build_select_from_with_joins(report.select_from, all_joins)

        return StreamingExportQuery(
            select_from=select_from,
            fields=selected_fields,
            conditions=conditions,
            orders=orders,
            max_rows=max_rows,
            statement_timeout_sec=statement_timeout_sec,
        )

    def build_session_query(
        self,
        report: ReportDef,
        fields: list[str] | None,
        filter: SessionExportFilter | None,
        order: list[SessionExportOrder] | None,
        max_rows: int,
        statement_timeout_sec: int,
    ) -> StreamingExportQuery:
        """Build StreamingExportQuery for session export.

        Dynamically applies JOINs based on selected fields.
        Fields with joins attribute will trigger LEFT JOIN operations.

        Args:
            report: Session report definition
            fields: Field keys to include (None = all fields)
            filter: Session-specific filter conditions
            order: Session-specific order specifications
            max_rows: Maximum rows to export
            statement_timeout_sec: Query timeout in seconds

        Returns:
            StreamingExportQuery ready for execution
        """
        selected_fields = self._select_fields(report, fields)
        conditions = self._build_session_conditions(report, filter)
        orders = self._build_session_orders(report, order)

        # Collect all required JOINs from selected fields
        all_joins = self._collect_joins(selected_fields)

        # Build select_from with dynamic JOINs
        select_from = self._build_select_from_with_joins(report.select_from, all_joins)

        return StreamingExportQuery(
            select_from=select_from,
            fields=selected_fields,
            conditions=conditions,
            orders=orders,
            max_rows=max_rows,
            statement_timeout_sec=statement_timeout_sec,
        )

    def build_project_query(
        self,
        report: ReportDef,
        fields: list[str] | None,
        filter: ProjectExportFilter | None,
        order: list[ProjectExportOrder] | None,
        max_rows: int,
        statement_timeout_sec: int,
    ) -> StreamingExportQuery:
        """Build StreamingExportQuery for project export.

        Dynamically applies JOINs based on selected fields.
        Fields with joins attribute will trigger LEFT JOIN operations.

        Args:
            report: Project report definition
            fields: Field keys to include (None = all fields)
            filter: Project-specific filter conditions
            order: Project-specific order specifications
            max_rows: Maximum rows to export
            statement_timeout_sec: Query timeout in seconds

        Returns:
            StreamingExportQuery ready for execution
        """
        selected_fields = self._select_fields(report, fields)
        conditions = self._build_project_conditions(report, filter)
        orders = self._build_project_orders(report, order)

        # Collect all required JOINs from selected fields
        all_joins = self._collect_joins(selected_fields)

        # Build select_from with dynamic JOINs
        select_from = self._build_select_from_with_joins(report.select_from, all_joins)

        return StreamingExportQuery(
            select_from=select_from,
            fields=selected_fields,
            conditions=conditions,
            orders=orders,
            max_rows=max_rows,
            statement_timeout_sec=statement_timeout_sec,
        )

    def build_keypair_query(
        self,
        report: ReportDef,
        fields: list[str] | None,
        filter: None,  # No filter support for keypair export yet
        order: None,  # No order support for keypair export yet
        max_rows: int,
        statement_timeout_sec: int,
    ) -> StreamingExportQuery:
        """Build StreamingExportQuery for keypair export.

        Dynamically applies JOINs based on selected fields.
        Fields with joins attribute will trigger LEFT JOIN operations.

        Args:
            report: Keypair report definition
            fields: Field keys to include (None = all fields)
            filter: Not yet supported for keypair export
            order: Not yet supported for keypair export
            max_rows: Maximum rows to export
            statement_timeout_sec: Query timeout in seconds

        Returns:
            StreamingExportQuery ready for execution
        """
        selected_fields = self._select_fields(report, fields)

        # Collect all required JOINs from selected fields
        all_joins = self._collect_joins(selected_fields)

        # Build select_from with dynamic JOINs
        select_from = self._build_select_from_with_joins(report.select_from, all_joins)

        return StreamingExportQuery(
            select_from=select_from,
            fields=selected_fields,
            conditions=[],  # No filter support yet
            orders=[],  # No order support yet
            max_rows=max_rows,
            statement_timeout_sec=statement_timeout_sec,
        )

    def _collect_joins(
        self,
        fields: list[ExportFieldDef],
    ) -> list[JoinDef]:
        """Collect all required JOINs from selected fields.

        Deduplicates JOINs and returns them in a consistent order
        for deterministic query generation.

        Args:
            fields: Selected field definitions

        Returns:
            List of unique JoinDef in deterministic order
        """
        seen: set[JoinDef] = set()
        result: list[JoinDef] = []

        for field in fields:
            if field.joins:
                for join_def in field.joins:
                    if join_def not in seen:
                        seen.add(join_def)
                        result.append(join_def)

        return result

    def _build_select_from_with_joins(
        self,
        base_table: sa.FromClause,
        joins: list[JoinDef],
    ) -> sa.FromClause:
        """Build select_from clause with dynamic LEFT JOINs.

        Args:
            base_table: Base table (e.g., GroupRow.__table__)
            joins: List of JoinDef to apply

        Returns:
            SQLAlchemy FromClause with all JOINs applied
        """
        result: sa.FromClause = base_table
        for join_def in joins:
            result = result.outerjoin(join_def.table, join_def.condition)
        return result

    def _select_fields(
        self,
        report: ReportDef,
        field_keys: list[str] | None,
    ) -> list[ExportFieldDef]:
        """Select fields from report based on requested field keys.

        Preserves the order specified in field_keys.

        Raises:
            InvalidExportFieldKeys: If a field key is not found in the report.
        """
        if field_keys:
            field_map = {f.key: f for f in report.fields}
            selected: list[ExportFieldDef] = []
            for key in field_keys:
                if key not in field_map:
                    raise InvalidExportFieldKeys([key])
                selected.append(field_map[key])
            return selected
        return list(report.fields)

    # =========================================================================
    # User-specific conditions and orders
    # =========================================================================

    def _build_user_conditions(
        self,
        report: ReportDef,
        filter: UserExportFilter | None,
    ) -> list[QueryCondition]:
        """Convert UserExportFilter to list of QueryCondition."""
        if filter is None:
            return []

        conditions: list[QueryCondition] = []

        # username filter
        if filter.username is not None:
            field = report.get_field("username")
            if field:
                cond = self._build_string_condition(filter.username, field)
                if cond:
                    conditions.append(cond)

        # email filter
        if filter.email is not None:
            field = report.get_field("email")
            if field:
                cond = self._build_string_condition(filter.email, field)
                if cond:
                    conditions.append(cond)

        # domain_name filter
        if filter.domain_name is not None:
            field = report.get_field("domain_name")
            if field:
                cond = self._build_string_condition(filter.domain_name, field)
                if cond:
                    conditions.append(cond)

        # role filter (IN query)
        if filter.role is not None:
            field = report.get_field("role")
            if field:
                cond = self._build_in_condition(filter.role, field)
                conditions.append(cond)

        # status filter (IN query)
        if filter.status is not None:
            field = report.get_field("status")
            if field:
                cond = self._build_in_condition(filter.status, field)
                conditions.append(cond)

        # created_at filter
        if filter.created_at is not None:
            field = report.get_field("created_at")
            if field:
                conditions.extend(self._build_datetime_conditions(filter.created_at, field))

        return conditions

    def _build_user_orders(
        self,
        report: ReportDef,
        orders: list[UserExportOrder] | None,
    ) -> list[QueryOrder]:
        """Convert UserExportOrder list to QueryOrder list."""
        if not orders:
            return []

        result: list[QueryOrder] = []
        for order in orders:
            field_def = report.get_field(order.field.value)
            if field_def is None:
                continue
            if order.direction == OrderDirection.ASC:
                result.append(field_def.column.asc())
            else:
                result.append(field_def.column.desc())

        return result

    # =========================================================================
    # Session-specific conditions and orders
    # =========================================================================

    def _build_session_conditions(
        self,
        report: ReportDef,
        filter: SessionExportFilter | None,
    ) -> list[QueryCondition]:
        """Convert SessionExportFilter to list of QueryCondition."""
        if filter is None:
            return []

        conditions: list[QueryCondition] = []

        # name filter
        if filter.name is not None:
            field = report.get_field("name")
            if field:
                cond = self._build_string_condition(filter.name, field)
                if cond:
                    conditions.append(cond)

        # session_type filter (IN query)
        if filter.session_type is not None:
            field = report.get_field("session_type")
            if field:
                cond = self._build_in_condition(filter.session_type, field)
                conditions.append(cond)

        # domain_name filter
        if filter.domain_name is not None:
            field = report.get_field("domain_name")
            if field:
                cond = self._build_string_condition(filter.domain_name, field)
                if cond:
                    conditions.append(cond)

        # access_key filter
        if filter.access_key is not None:
            field = report.get_field("access_key")
            if field:
                cond = self._build_string_condition(filter.access_key, field)
                if cond:
                    conditions.append(cond)

        # status filter (IN query)
        if filter.status is not None:
            field = report.get_field("status")
            if field:
                cond = self._build_in_condition(filter.status, field)
                conditions.append(cond)

        # scaling_group_name filter
        if filter.scaling_group_name is not None:
            field = report.get_field("scaling_group_name")
            if field:
                cond = self._build_string_condition(filter.scaling_group_name, field)
                if cond:
                    conditions.append(cond)

        # created_at filter
        if filter.created_at is not None:
            field = report.get_field("created_at")
            if field:
                conditions.extend(self._build_datetime_conditions(filter.created_at, field))

        # terminated_at filter
        if filter.terminated_at is not None:
            field = report.get_field("terminated_at")
            if field:
                conditions.extend(self._build_datetime_conditions(filter.terminated_at, field))

        return conditions

    def _build_session_orders(
        self,
        report: ReportDef,
        orders: list[SessionExportOrder] | None,
    ) -> list[QueryOrder]:
        """Convert SessionExportOrder list to QueryOrder list."""
        if not orders:
            return []

        result: list[QueryOrder] = []
        for order in orders:
            field_def = report.get_field(order.field.value)
            if field_def is None:
                continue
            if order.direction == OrderDirection.ASC:
                result.append(field_def.column.asc())
            else:
                result.append(field_def.column.desc())

        return result

    # =========================================================================
    # Project-specific conditions and orders
    # =========================================================================

    def _build_project_conditions(
        self,
        report: ReportDef,
        filter: ProjectExportFilter | None,
    ) -> list[QueryCondition]:
        """Convert ProjectExportFilter to list of QueryCondition."""
        if filter is None:
            return []

        conditions: list[QueryCondition] = []

        # name filter
        if filter.name is not None:
            field = report.get_field("name")
            if field:
                cond = self._build_string_condition(filter.name, field)
                if cond:
                    conditions.append(cond)

        # domain_name filter
        if filter.domain_name is not None:
            field = report.get_field("domain_name")
            if field:
                cond = self._build_string_condition(filter.domain_name, field)
                if cond:
                    conditions.append(cond)

        # is_active filter (boolean)
        if filter.is_active is not None:
            field = report.get_field("is_active")
            if field:
                conditions.append(self._build_boolean_condition(filter.is_active, field))

        # created_at filter
        if filter.created_at is not None:
            field = report.get_field("created_at")
            if field:
                conditions.extend(self._build_datetime_conditions(filter.created_at, field))

        return conditions

    def _build_project_orders(
        self,
        report: ReportDef,
        orders: list[ProjectExportOrder] | None,
    ) -> list[QueryOrder]:
        """Convert ProjectExportOrder list to QueryOrder list."""
        if not orders:
            return []

        result: list[QueryOrder] = []
        for order in orders:
            field_def = report.get_field(order.field.value)
            if field_def is None:
                continue
            if order.direction == OrderDirection.ASC:
                result.append(field_def.column.asc())
            else:
                result.append(field_def.column.desc())

        return result

    # =========================================================================
    # Audit Log-specific query builder, conditions and orders
    # =========================================================================

    def build_audit_log_query(
        self,
        report: ReportDef,
        fields: list[str] | None,
        filter: AuditLogExportFilter | None,
        order: list[AuditLogExportOrder] | None,
        max_rows: int,
        statement_timeout_sec: int,
    ) -> StreamingExportQuery:
        """Build StreamingExportQuery for audit log export.

        Args:
            report: Audit log report definition
            fields: Field keys to include (None = all fields)
            filter: Audit log-specific filter conditions
            order: Audit log-specific order specifications
            max_rows: Maximum rows to export
            statement_timeout_sec: Query timeout in seconds

        Returns:
            StreamingExportQuery ready for execution
        """
        selected_fields = self._select_fields(report, fields)
        conditions = self._build_audit_log_conditions(report, filter)
        orders = self._build_audit_log_orders(report, order)

        return StreamingExportQuery(
            select_from=report.select_from,
            fields=selected_fields,
            conditions=conditions,
            orders=orders,
            max_rows=max_rows,
            statement_timeout_sec=statement_timeout_sec,
        )

    def _build_audit_log_conditions(
        self,
        report: ReportDef,
        filter: AuditLogExportFilter | None,
    ) -> list[QueryCondition]:
        """Convert AuditLogExportFilter to list of QueryCondition."""
        if filter is None:
            return []

        conditions: list[QueryCondition] = []

        # entity_type filter
        if filter.entity_type is not None:
            field = report.get_field("entity_type")
            if field:
                cond = self._build_string_condition(filter.entity_type, field)
                if cond:
                    conditions.append(cond)

        # entity_id filter
        if filter.entity_id is not None:
            field = report.get_field("entity_id")
            if field:
                cond = self._build_string_condition(filter.entity_id, field)
                if cond:
                    conditions.append(cond)

        # operation filter
        if filter.operation is not None:
            field = report.get_field("operation")
            if field:
                cond = self._build_string_condition(filter.operation, field)
                if cond:
                    conditions.append(cond)

        # status filter (IN query)
        if filter.status is not None:
            field = report.get_field("status")
            if field:
                cond = self._build_in_condition(filter.status, field)
                conditions.append(cond)

        # triggered_by filter
        if filter.triggered_by is not None:
            field = report.get_field("triggered_by")
            if field:
                cond = self._build_string_condition(filter.triggered_by, field)
                if cond:
                    conditions.append(cond)

        # request_id filter
        if filter.request_id is not None:
            field = report.get_field("request_id")
            if field:
                cond = self._build_string_condition(filter.request_id, field)
                if cond:
                    conditions.append(cond)

        # created_at filter
        if filter.created_at is not None:
            field = report.get_field("created_at")
            if field:
                conditions.extend(self._build_datetime_conditions(filter.created_at, field))

        return conditions

    def _build_audit_log_orders(
        self,
        report: ReportDef,
        orders: list[AuditLogExportOrder] | None,
    ) -> list[QueryOrder]:
        """Convert AuditLogExportOrder list to QueryOrder list."""
        if not orders:
            return []

        result: list[QueryOrder] = []
        for order in orders:
            field_def = report.get_field(order.field.value)
            if field_def is None:
                continue
            if order.direction == OrderDirection.ASC:
                result.append(field_def.column.asc())
            else:
                result.append(field_def.column.desc())

        return result

    # =========================================================================
    # Common filter builders
    # =========================================================================

    def _build_in_condition(
        self,
        values: list[str],
        field_def: ExportFieldDef,
    ) -> QueryCondition:
        """Build IN condition for a field.

        If the field type is ENUM, converts string values to enum values.
        """
        column = field_def.column

        if field_def.field_type == ExportFieldType.ENUM:
            enum_cls = column.type._enum_cls
            enum_values = [enum_cls(v) for v in values]
            return lambda: column.in_(enum_values)

        return lambda: column.in_(values)

    def _build_string_condition(
        self,
        string_filter: StringFilter,
        field_def: ExportFieldDef,
    ) -> QueryCondition | None:
        """Convert StringFilter to QueryCondition for the given field."""
        column = field_def.column

        def make_contains_factory(
            col: InstrumentedAttribute[Any],
        ) -> Callable[[StringMatchSpec], QueryCondition]:
            def factory(spec: StringMatchSpec) -> QueryCondition:
                pattern = f"%{spec.value}%"
                if spec.case_insensitive:
                    if spec.negated:
                        return lambda: ~col.ilike(pattern)
                    return lambda: col.ilike(pattern)
                if spec.negated:
                    return lambda: ~col.like(pattern)
                return lambda: col.like(pattern)

            return factory

        def make_equals_factory(
            col: InstrumentedAttribute[Any],
        ) -> Callable[[StringMatchSpec], QueryCondition]:
            def factory(spec: StringMatchSpec) -> QueryCondition:
                if spec.case_insensitive:
                    if spec.negated:
                        return lambda: ~col.ilike(spec.value)
                    return lambda: col.ilike(spec.value)
                if spec.negated:
                    return lambda: col != spec.value
                return lambda: col == spec.value

            return factory

        def make_starts_with_factory(
            col: InstrumentedAttribute[Any],
        ) -> Callable[[StringMatchSpec], QueryCondition]:
            def factory(spec: StringMatchSpec) -> QueryCondition:
                pattern = f"{spec.value}%"
                if spec.case_insensitive:
                    if spec.negated:
                        return lambda: ~col.ilike(pattern)
                    return lambda: col.ilike(pattern)
                if spec.negated:
                    return lambda: ~col.like(pattern)
                return lambda: col.like(pattern)

            return factory

        def make_ends_with_factory(
            col: InstrumentedAttribute[Any],
        ) -> Callable[[StringMatchSpec], QueryCondition]:
            def factory(spec: StringMatchSpec) -> QueryCondition:
                pattern = f"%{spec.value}"
                if spec.case_insensitive:
                    if spec.negated:
                        return lambda: ~col.ilike(pattern)
                    return lambda: col.ilike(pattern)
                if spec.negated:
                    return lambda: ~col.like(pattern)
                return lambda: col.like(pattern)

            return factory

        return self.convert_string_filter(
            string_filter,
            contains_factory=make_contains_factory(column),
            equals_factory=make_equals_factory(column),
            starts_with_factory=make_starts_with_factory(column),
            ends_with_factory=make_ends_with_factory(column),
        )

    def _build_datetime_conditions(
        self,
        dt_filter: DateTimeRangeFilter,
        field_def: ExportFieldDef,
    ) -> list[QueryCondition]:
        """Convert DateTimeRangeFilter to list of QueryConditions."""
        conditions: list[QueryCondition] = []
        column = field_def.column

        if dt_filter.after is not None:
            after_dt = dt_filter.after
            conditions.append(lambda: column >= after_dt)

        if dt_filter.before is not None:
            before_dt = dt_filter.before
            conditions.append(lambda: column <= before_dt)

        return conditions

    def _build_boolean_condition(
        self,
        bool_filter: BooleanFilter,
        field_def: ExportFieldDef,
    ) -> QueryCondition:
        """Convert BooleanFilter to QueryCondition."""
        column = field_def.column
        value = bool_filter.equals
        return lambda: column == value



"""Utility functions for resource preset repository."""

from __future__ import annotations

import logging
from collections.abc import Generator
from contextlib import contextmanager

from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@contextmanager
def suppress_with_log(
    exceptions: list[type[BaseException]],
    message: str | None = None,
    log_level: int = logging.WARNING,
) -> Generator[None, None, None]:
    """
    Context manager that suppresses specified exceptions and logs them.

    Args:
        exceptions: List of exception types to suppress
        message: Optional custom message to log with the exception
        log_level: Logging level to use (default: WARNING)
    """
    try:
        yield
    except tuple(exceptions) as e:
        if message:
            log.log(log_level, "{}: {}", message, e)
        else:
            log.log(log_level, "Suppressed exception: {}", e)



"""Route-specific recorder context for route coordinator."""

from __future__ import annotations

from uuid import UUID

from ai.backend.manager.sokovan.recorder.context import RecorderContext

# Route-specific type alias
RouteRecorderContext = RecorderContext[UUID]

__all__ = [
    "RouteRecorderContext",
]



"""Fair Share Repository with Resilience policies."""

from __future__ import annotations

import uuid
from collections.abc import Sequence
from datetime import date
from typing import TYPE_CHECKING

from ai.backend.common.metrics.metric import DomainType, LayerType
from ai.backend.common.resilience import (
    MetricArgs,
    MetricPolicy,
    Resilience,
    RetryArgs,
    RetryPolicy,
)
from ai.backend.common.resilience.policies.retry import BackoffStrategy
from ai.backend.manager.data.fair_share import (
    DomainFairShareData,
    DomainFairShareSearchResult,
    FairShareCalculationContext,
    ProjectFairShareData,
    ProjectFairShareSearchResult,
    ProjectUserIds,
    UserFairShareData,
    UserFairShareFactors,
    UserFairShareSearchResult,
)
from ai.backend.manager.models.scaling_group.types import FairShareScalingGroupSpec
from ai.backend.manager.repositories.base import (
    BatchQuerier,
    BulkUpserter,
    BulkUpserterResult,
    Creator,
    Upserter,
)

from .db_source import FairShareDBSource
from .types import (
    DomainFairShareEntitySearchResult,
    DomainFairShareSearchScope,
    ProjectFairShareEntitySearchResult,
    ProjectFairShareSearchScope,
    UserFairShareEntitySearchResult,
    UserFairShareSearchScope,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.fair_share import (
        DomainFairShareRow,
        ProjectFairShareRow,
        UserFairShareRow,
    )
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine
    from ai.backend.manager.sokovan.scheduler.fair_share import FairShareFactorCalculationResult


__all__ = ("FairShareRepository",)


fair_share_repository_resilience = Resilience(
    policies=[
        MetricPolicy(
            MetricArgs(domain=DomainType.REPOSITORY, layer=LayerType.FAIR_SHARE_REPOSITORY)
        ),
        RetryPolicy(
            RetryArgs(
                max_retries=10,
                retry_delay=0.1,
                backoff_strategy=BackoffStrategy.FIXED,
            )
        ),
    ]
)


class FairShareRepository:
    """Repository for Fair Share data access with resilience policies."""

    _db_source: FairShareDBSource

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db_source = FairShareDBSource(db)

    # ==================== Domain Fair Share ====================

    @fair_share_repository_resilience.apply()
    async def create_domain_fair_share(
        self,
        creator: Creator[DomainFairShareRow],
    ) -> DomainFairShareData:
        """Create a new domain fair share record."""
        return await self._db_source.create_domain_fair_share(creator)

    @fair_share_repository_resilience.apply()
    async def upsert_domain_fair_share(
        self,
        upserter: Upserter[DomainFairShareRow],
    ) -> DomainFairShareData:
        """Upsert a domain fair share record."""
        return await self._db_source.upsert_domain_fair_share(upserter)

    @fair_share_repository_resilience.apply()
    async def get_domain_fair_share(
        self,
        resource_group: str,
        domain_name: str,
    ) -> DomainFairShareData:
        """Get a domain fair share record by scaling group and domain name.

        Returns default values if the domain exists but has no fair share record.

        Raises:
            DomainNotFound: If the domain does not exist.
        """
        return await self._db_source.get_domain_fair_share(resource_group, domain_name)

    @fair_share_repository_resilience.apply()
    async def search_domain_fair_shares(
        self,
        querier: BatchQuerier,
    ) -> DomainFairShareSearchResult:
        """Search domain fair shares with pagination."""
        return await self._db_source.search_domain_fair_shares(querier)

    @fair_share_repository_resilience.apply()
    async def search_rg_domain_fair_shares(
        self,
        scope: DomainFairShareSearchScope,
        querier: BatchQuerier,
    ) -> DomainFairShareEntitySearchResult:
        """Search domain fair shares within a resource group.

        Returns all domains associated with a resource group with complete fair share data.

        Args:
            scope: Required scope with resource_group.
            querier: Pagination, conditions, and orders for the query.

        Returns:
            DomainFairShareEntitySearchResult with domain entities and their complete details.
        """
        return await self._db_source.search_rg_domain_fair_shares(scope, querier)

    # ==================== Project Fair Share ====================

    @fair_share_repository_resilience.apply()
    async def create_project_fair_share(
        self,
        creator: Creator[ProjectFairShareRow],
    ) -> ProjectFairShareData:
        """Create a new project fair share record."""
        return await self._db_source.create_project_fair_share(creator)

    @fair_share_repository_resilience.apply()
    async def upsert_project_fair_share(
        self,
        upserter: Upserter[ProjectFairShareRow],
    ) -> ProjectFairShareData:
        """Upsert a project fair share record."""
        return await self._db_source.upsert_project_fair_share(upserter)

    @fair_share_repository_resilience.apply()
    async def get_project_fair_share(
        self,
        resource_group: str,
        project_id: uuid.UUID,
    ) -> ProjectFairShareData:
        """Get a project fair share record by scaling group and project ID.

        Returns default values if the project exists but has no fair share record.

        Raises:
            ProjectNotFound: If the project does not exist.
        """
        return await self._db_source.get_project_fair_share(resource_group, project_id)

    @fair_share_repository_resilience.apply()
    async def search_project_fair_shares(
        self,
        querier: BatchQuerier,
    ) -> ProjectFairShareSearchResult:
        """Search project fair shares with pagination."""
        return await self._db_source.search_project_fair_shares(querier)

    @fair_share_repository_resilience.apply()
    async def search_rg_project_fair_shares(
        self,
        scope: ProjectFairShareSearchScope,
        querier: BatchQuerier,
    ) -> ProjectFairShareEntitySearchResult:
        """Search project fair shares within a resource group.

        Returns all projects associated with a resource group with complete fair share data.

        Args:
            scope: Required scope with resource_group.
            querier: Pagination, conditions, and orders for the query.

        Returns:
            ProjectFairShareEntitySearchResult with project entities and their complete details.
        """
        return await self._db_source.search_rg_project_fair_shares(scope, querier)

    # ==================== User Fair Share ====================

    @fair_share_repository_resilience.apply()
    async def create_user_fair_share(
        self,
        creator: Creator[UserFairShareRow],
    ) -> UserFairShareData:
        """Create a new user fair share record."""
        return await self._db_source.create_user_fair_share(creator)

    @fair_share_repository_resilience.apply()
    async def upsert_user_fair_share(
        self,
        upserter: Upserter[UserFairShareRow],
    ) -> UserFairShareData:
        """Upsert a user fair share record."""
        return await self._db_source.upsert_user_fair_share(upserter)

    # ==================== Bulk Upsert Operations ====================

    @fair_share_repository_resilience.apply()
    async def bulk_upsert_domain_fair_share(
        self,
        bulk_upserter: BulkUpserter[DomainFairShareRow],
    ) -> BulkUpserterResult:
        """Bulk upsert domain fair share records."""
        return await self._db_source.bulk_upsert_domain_fair_share(bulk_upserter)

    @fair_share_repository_resilience.apply()
    async def bulk_upsert_project_fair_share(
        self,
        bulk_upserter: BulkUpserter[ProjectFairShareRow],
    ) -> BulkUpserterResult:
        """Bulk upsert project fair share records."""
        return await self._db_source.bulk_upsert_project_fair_share(bulk_upserter)

    @fair_share_repository_resilience.apply()
    async def bulk_upsert_user_fair_share(
        self,
        bulk_upserter: BulkUpserter[UserFairShareRow],
    ) -> BulkUpserterResult:
        """Bulk upsert user fair share records."""
        return await self._db_source.bulk_upsert_user_fair_share(bulk_upserter)

    @fair_share_repository_resilience.apply()
    async def get_user_fair_share(
        self,
        resource_group: str,
        project_id: uuid.UUID,
        user_uuid: uuid.UUID,
    ) -> UserFairShareData:
        """Get a user fair share record by scaling group, project ID, and user UUID.

        Returns default values if the user exists in the project but has no fair share record.

        Raises:
            UserNotFound: If the user does not exist in the project.
        """
        return await self._db_source.get_user_fair_share(resource_group, project_id, user_uuid)

    @fair_share_repository_resilience.apply()
    async def search_user_fair_shares(
        self,
        querier: BatchQuerier,
    ) -> UserFairShareSearchResult:
        """Search user fair shares with pagination."""
        return await self._db_source.search_user_fair_shares(querier)

    @fair_share_repository_resilience.apply()
    async def search_rg_user_fair_shares(
        self,
        scope: UserFairShareSearchScope,
        querier: BatchQuerier,
    ) -> UserFairShareEntitySearchResult:
        """Search user fair shares within a resource group.

        Returns all users associated with a resource group (via project membership)
        with complete fair share data.

        Args:
            scope: Required scope with resource_group.
            querier: Pagination, conditions, and orders for the query.

        Returns:
            UserFairShareEntitySearchResult with user entities and their complete details.
        """
        return await self._db_source.search_rg_user_fair_shares(scope, querier)

    # ==================== Entity Info & Spec ====================

    @fair_share_repository_resilience.apply()
    async def get_user_project_info(
        self,
        project_id: uuid.UUID,
        user_uuid: uuid.UUID,
    ) -> str | None:
        """Get domain_name if user exists in project.

        Returns:
            domain_name if user is member of project, None otherwise.
        """
        return await self._db_source.get_user_project_info(project_id, user_uuid)

    @fair_share_repository_resilience.apply()
    async def get_project_info(
        self,
        project_id: uuid.UUID,
    ) -> str | None:
        """Get domain_name if project exists.

        Returns:
            domain_name if project exists, None otherwise.
        """
        return await self._db_source.get_project_info(project_id)

    @fair_share_repository_resilience.apply()
    async def get_domain_exists(
        self,
        domain_name: str,
    ) -> bool:
        """Check if domain exists.

        Returns:
            True if domain exists, False otherwise.
        """
        return await self._db_source.get_domain_exists(domain_name)

    @fair_share_repository_resilience.apply()
    async def get_scaling_group_fair_share_spec(
        self,
        scaling_group: str,
    ) -> FairShareScalingGroupSpec:
        """Get fair share spec for scaling group.

        Returns:
            FairShareScalingGroupSpec with defaults if not configured.

        Raises:
            ScalingGroupNotFound: If scaling group doesn't exist.
        """
        return await self._db_source.get_scaling_group_fair_share_spec(scaling_group)

    @fair_share_repository_resilience.apply()
    async def get_user_scheduling_ranks_batch(
        self,
        resource_group: str,
        project_user_ids: Sequence[ProjectUserIds],
    ) -> dict[uuid.UUID, int]:
        """Get scheduling ranks for multiple users across projects.

        Args:
            resource_group: The resource group (scaling group) name.
            project_user_ids: Sequence of ProjectUserIds containing project and user IDs.

        Returns:
            A mapping from user_uuid to scheduling_rank.
            Users not found in the database or with NULL rank are omitted.
        """
        return await self._db_source.get_user_scheduling_ranks_batch(
            resource_group, project_user_ids
        )

    # ==================== Bulk Factor Updates ====================

    @fair_share_repository_resilience.apply()
    async def bulk_update_fair_share_factors(
        self,
        resource_group: str,
        calculation_result: FairShareFactorCalculationResult,
        lookback_start: date,
        lookback_end: date,
    ) -> None:
        """Bulk update fair share factors for all levels.

        Updates domain, project, and user fair share records with calculated
        factors in a single transaction.

        Args:
            resource_group: The resource group being updated
            calculation_result: Calculated factors from FairShareFactorCalculator
            lookback_start: Start of lookback period used in calculation
            lookback_end: End of lookback period used in calculation
        """
        return await self._db_source.bulk_update_fair_share_factors(
            resource_group, calculation_result, lookback_start, lookback_end
        )

    # ==================== Batched Reads ====================

    @fair_share_repository_resilience.apply()
    async def get_user_fair_share_factors_batch(
        self,
        resource_group: str,
        project_user_ids: Sequence[ProjectUserIds],
    ) -> dict[uuid.UUID, UserFairShareFactors]:
        """Get combined fair share factors for multiple users with 3-way JOIN.

        Fetches domain, project, and user fair share factors in a single query
        by joining the three fair share tables. Used for factor-based workload
        sequencing in FairShareSequencer.

        Args:
            resource_group: The resource group (scaling group) name.
            project_user_ids: Sequence of ProjectUserIds containing project and user IDs.

        Returns:
            A mapping from user_uuid to UserFairShareFactors containing all three factor levels.
            Users not found in any of the fair share tables are omitted.
        """
        return await self._db_source.get_user_fair_share_factors_batch(
            resource_group, project_user_ids
        )

    @fair_share_repository_resilience.apply()
    async def get_fair_share_calculation_context(
        self,
        scaling_group: str,
        today: date,
    ) -> FairShareCalculationContext:
        """Get all data needed for fair share factor calculation.

        Fetches scaling group config, fair share records, and decayed usages
        in one database session for consistency and efficiency.

        Args:
            scaling_group: The scaling group name
            today: Current date for decay calculation

        Returns:
            FairShareCalculationContext containing all data for factor calculation

        Raises:
            ScalingGroupNotFound: If the scaling group doesn't exist
        """
        return await self._db_source.get_fair_share_calculation_context(scaling_group, today)



from __future__ import annotations

import uuid
from collections.abc import Sequence

from ai.backend.manager.api.gql.base import UUIDInMatchSpec
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.repositories.base import BatchQuerier, NoPagination
from ai.backend.manager.repositories.group.options import GroupConditions
from ai.backend.manager.services.group.actions.search_projects import SearchProjectsAction
from ai.backend.manager.services.group.processors import GroupProcessors


async def load_projects_by_ids(
    processor: GroupProcessors,
    project_ids: Sequence[uuid.UUID],
) -> list[GroupData | None]:
    """Batch load projects by their IDs.

    Args:
        processor: The group processor.
        project_ids: Sequence of project UUIDs to load.

    Returns:
        List of GroupData (or None if not found) in the same order as project_ids.
    """
    if not project_ids:
        return []

    querier = BatchQuerier(
        pagination=NoPagination(),
        conditions=[
            GroupConditions.by_id_in(UUIDInMatchSpec(values=list(project_ids), negated=False))
        ],
    )

    action_result = await processor.search_projects.wait_for_complete(
        SearchProjectsAction(querier=querier)
    )

    project_map = {group.id: group for group in action_result.items}
    return [project_map.get(project_id) for project_id in project_ids]



import uuid
from dataclasses import dataclass
from typing import override

from ai.backend.common.data.artifact.types import ArtifactRegistryType
from ai.backend.common.data.permission.types import EntityType
from ai.backend.common.data.storage.registries.types import ModelSortKey
from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.artifact.types import (
    ArtifactDataWithRevisions,
    ArtifactRevisionReadme,
    ArtifactType,
    DelegateeTarget,
)
from ai.backend.manager.services.artifact.actions.base import ArtifactAction


@dataclass
class DelegateScanArtifactsAction(ArtifactAction):
    delegator_reservoir_id: uuid.UUID | None
    artifact_type: ArtifactType | None
    search: str | None
    order: ModelSortKey | None
    delegatee_target: DelegateeTarget | None
    limit: int | None

    @override
    @classmethod
    def entity_type(cls) -> EntityType:
        return EntityType.ARTIFACT_SCAN

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.CREATE


@dataclass
class DelegateScanArtifactsActionResult(BaseActionResult):
    result: list[ArtifactDataWithRevisions]
    source_registry_id: uuid.UUID
    source_registry_type: ArtifactRegistryType
    readme_data: dict[uuid.UUID, ArtifactRevisionReadme]

    @override
    def entity_id(self) -> str | None:
        return None



from __future__ import annotations

import decimal
import logging

from ai.backend.common.contexts.user import current_user
from ai.backend.logging.utils import BraceStyleAdapter
from ai.backend.manager.errors.service import EndpointAccessForbiddenError
from ai.backend.manager.models.user import UserRole
from ai.backend.manager.repositories.model_serving.options import EndpointConditions
from ai.backend.manager.repositories.model_serving.repository import ModelServingRepository
from ai.backend.manager.services.model_serving.actions.create_auto_scaling_rule import (
    CreateEndpointAutoScalingRuleAction,
    CreateEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.delete_auto_scaling_rule import (
    DeleteEndpointAutoScalingRuleAction,
    DeleteEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.modify_auto_scaling_rule import (
    ModifyEndpointAutoScalingRuleAction,
    ModifyEndpointAutoScalingRuleActionResult,
)
from ai.backend.manager.services.model_serving.actions.scale_service_replicas import (
    ScaleServiceReplicasAction,
    ScaleServiceReplicasActionResult,
)
from ai.backend.manager.services.model_serving.actions.search_auto_scaling_rules import (
    SearchAutoScalingRulesAction,
    SearchAutoScalingRulesActionResult,
)
from ai.backend.manager.services.model_serving.exceptions import (
    EndpointAutoScalingRuleNotFound,
    EndpointNotFound,
    GenericForbidden,
    InvalidAPIParameters,
    ModelServiceNotFound,
)
from ai.backend.manager.services.model_serving.services.utils import validate_endpoint_access

log = BraceStyleAdapter(logging.getLogger(__name__))


class AutoScalingService:
    _repository: ModelServingRepository

    def __init__(
        self,
        repository: ModelServingRepository,
    ) -> None:
        self._repository = repository

    async def check_user_access(self) -> None:
        user_data = current_user()
        if user_data is None or user_data.is_authorized is False:
            raise GenericForbidden("Only authorized requests may have access key scopes.")

    async def scale_service_replicas(
        self, action: ScaleServiceReplicasAction
    ) -> ScaleServiceReplicasActionResult:
        # Validate access
        await self.check_user_access()
        validation_data = await self._repository.get_endpoint_access_validation_data(
            action.service_id
        )
        if not validation_data:
            raise ModelServiceNotFound
        if not validate_endpoint_access(validation_data):
            raise EndpointAccessForbiddenError

        # Get endpoint data
        endpoint_data = await self._repository.get_endpoint_by_id(action.service_id)
        if not endpoint_data:
            raise ModelServiceNotFound

        # Update replicas (access already validated)
        success = await self._repository.update_endpoint_replicas(action.service_id, action.to)
        if not success:
            raise ModelServiceNotFound

        return ScaleServiceReplicasActionResult(
            current_route_count=len(endpoint_data.routings) if endpoint_data.routings else 0,
            target_count=action.to,
        )

    async def create_endpoint_auto_scaling_rule(
        self, action: CreateEndpointAutoScalingRuleAction
    ) -> CreateEndpointAutoScalingRuleActionResult:
        try:
            _threshold = decimal.Decimal(action.creator.threshold)
        except decimal.InvalidOperation as e:
            raise InvalidAPIParameters(
                f"Cannot convert {action.creator.threshold} to Decimal"
            ) from e

        # Validate access to the endpoint first
        validation_data = await self._repository.get_endpoint_access_validation_data(
            action.endpoint_id
        )
        if not validation_data:
            raise EndpointNotFound
        if not validate_endpoint_access(validation_data):
            raise EndpointAccessForbiddenError

        # Create auto scaling rule (access already validated)
        created_rule = await self._repository.create_auto_scaling_rule(
            endpoint_id=action.endpoint_id,
            metric_source=action.creator.metric_source,
            metric_name=action.creator.metric_name,
            threshold=_threshold,
            comparator=action.creator.comparator,
            step_size=action.creator.step_size,
            cooldown_seconds=action.creator.cooldown_seconds,
            min_replicas=action.creator.min_replicas,
            max_replicas=action.creator.max_replicas,
        )

        if created_rule is None:
            raise EndpointNotFound

        return CreateEndpointAutoScalingRuleActionResult(
            success=True,
            data=created_rule,
        )

    async def modify_endpoint_auto_scaling_rule(
        self, action: ModifyEndpointAutoScalingRuleAction
    ) -> ModifyEndpointAutoScalingRuleActionResult:
        # Get the rule to find the endpoint
        rule_data = await self._repository.get_auto_scaling_rule_by_id(action.id)
        if rule_data is None:
            raise EndpointAutoScalingRuleNotFound

        # Validate access to the endpoint
        validation_data = await self._repository.get_endpoint_access_validation_data(
            rule_data.endpoint
        )
        if not validation_data:
            raise EndpointNotFound
        if not validate_endpoint_access(validation_data):
            raise EndpointAccessForbiddenError

        # Update auto scaling rule (access already validated)
        updated_rule = await self._repository.update_auto_scaling_rule(action.updater)
        if updated_rule is None:
            raise EndpointAutoScalingRuleNotFound

        return ModifyEndpointAutoScalingRuleActionResult(
            success=True,
            data=updated_rule,
        )

    async def delete_endpoint_auto_scaling_rule(
        self, action: DeleteEndpointAutoScalingRuleAction
    ) -> DeleteEndpointAutoScalingRuleActionResult:
        # Get the rule to find the endpoint
        rule_data = await self._repository.get_auto_scaling_rule_by_id(action.id)
        if rule_data is None:
            raise EndpointAutoScalingRuleNotFound

        # Validate access to the endpoint
        validation_data = await self._repository.get_endpoint_access_validation_data(
            rule_data.endpoint
        )
        if not validation_data:
            raise EndpointNotFound
        if not validate_endpoint_access(validation_data):
            raise EndpointAccessForbiddenError

        # Delete auto scaling rule (access already validated)
        success = await self._repository.delete_auto_scaling_rule(action.id)
        if not success:
            raise EndpointAutoScalingRuleNotFound

        return DeleteEndpointAutoScalingRuleActionResult(
            success=True,
        )

    async def search_auto_scaling_rules(
        self, action: SearchAutoScalingRulesAction
    ) -> SearchAutoScalingRulesActionResult:
        """Searches endpoint auto scaling rules."""
        await self.check_user_access()

        # Apply access control conditions based on role
        user_data = current_user()
        if user_data is None:
            raise GenericForbidden("User context not available.")

        match user_data.role:
            case UserRole.SUPERADMIN | UserRole.MONITOR:
                pass  # No additional conditions for SUPERADMIN and MONITOR
            case UserRole.ADMIN:
                action.querier.conditions.append(
                    EndpointConditions.by_domain(user_data.domain_name)
                )
            case UserRole.USER:
                action.querier.conditions.append(
                    EndpointConditions.by_session_owner(user_data.user_id)
                )

        result = await self._repository.search_auto_scaling_rules(querier=action.querier)

        return SearchAutoScalingRulesActionResult(
            rules=result.items,
            total_count=result.total_count,
            has_next_page=result.has_next_page,
            has_previous_page=result.has_previous_page,
        )



import logging
import re
from collections.abc import Iterable
from typing import TYPE_CHECKING, Any

import aiohttp_cors
import sqlalchemy as sa
import trafaret as t
from aiohttp import web

from ai.backend.common import msgpack
from ai.backend.common import validators as tx
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.errors.api import InvalidAPIParameters
from ai.backend.manager.errors.common import GenericForbidden
from ai.backend.manager.errors.resource import ProjectNotFound
from ai.backend.manager.errors.storage import (
    DotfileAlreadyExists,
    DotfileCreationFailed,
    DotfileNotFound,
)
from ai.backend.manager.models.domain import MAXIMUM_DOTFILE_SIZE, verify_dotfile_name
from ai.backend.manager.models.group import (
    association_groups_users as agus,
)
from ai.backend.manager.models.group import (
    groups,
    query_group_domain,
    query_group_dotfiles,
)

from .auth import admin_required, auth_required
from .manager import READ_ALLOWED, server_status_required
from .types import CORSOptions, WebMiddleware
from .utils import check_api_params

if TYPE_CHECKING:
    from .context import RootContext

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["group", "groupId", "group_id"]): tx.UUID | t.String,
            t.Key("domain", default=None): t.String | t.Null,
            t.Key("data"): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
            t.Key("path"): t.String,
            t.Key("permission"): t.Regexp(r"^[0-7]{3}$", re.ASCII),
        },
    )
)
async def create(request: web.Request, params: Any) -> web.Response:
    log.info("GROUPCONFIG.CREATE_DOTFILE (group: {0})", params["group"])
    root_ctx: RootContext = request.app["_root.context"]
    group_id_or_name = params["group"]
    async with root_ctx.db.begin() as conn:
        if isinstance(group_id_or_name, str):
            if params["domain"] is None:
                raise InvalidAPIParameters("Missing parameter 'domain'")

            query = (
                sa.select(groups.c.id)
                .select_from(groups)
                .where(groups.c.domain_name == params["domain"])
                .where(groups.c.name == group_id_or_name)
            )
            group_id = await conn.scalar(query)
            domain = params["domain"]
        else:
            group_id = group_id_or_name
            # if group UUID is given, override input domain
            domain = await query_group_domain(conn, group_id)
        if group_id is None or domain is None:
            raise ProjectNotFound
        if not request["is_superadmin"] and request["user"]["domain_name"] != domain:
            raise GenericForbidden("Admins cannot create group dotfiles of other domains")

        dotfiles, leftover_space = await query_group_dotfiles(conn, group_id)
        if dotfiles is None:
            raise ProjectNotFound
        if leftover_space == 0:
            raise DotfileCreationFailed("No leftover space for dotfile storage")
        if len(dotfiles) == 100:
            raise DotfileCreationFailed("Dotfile creation limit reached")
        if not verify_dotfile_name(params["path"]):
            raise InvalidAPIParameters("dotfile path is reserved for internal operations.")

        duplicate = [x for x in dotfiles if x["path"] == params["path"]]
        if len(duplicate) > 0:
            raise DotfileAlreadyExists
        new_dotfiles = list(dotfiles)
        new_dotfiles.append({
            "path": params["path"],
            "perm": params["permission"],
            "data": params["data"],
        })
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed("No leftover space for dotfile storage")

        query = groups.update().values(dotfiles=dotfile_packed).where(groups.c.id == group_id)
        await conn.execute(query)
    return web.json_response({})


@server_status_required(READ_ALLOWED)
@auth_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["group", "groupId", "group_id"]): tx.UUID | t.String,
        t.Key("domain", default=None): t.String | t.Null,
        t.Key("path", default=None): t.Null | t.String,
    })
)
async def list_or_get(request: web.Request, params: Any) -> web.Response:
    log.info("GROUPCONFIG.LIST_OR_GET_DOTFILE (group: {0})", params["group"])
    root_ctx: RootContext = request.app["_root.context"]
    resp = []
    group_id_or_name = params["group"]
    async with root_ctx.db.begin() as conn:
        if isinstance(group_id_or_name, str):
            if params["domain"] is None:
                raise InvalidAPIParameters("Missing parameter 'domain'")
            query = (
                sa.select(groups.c.id)
                .select_from(groups)
                .where(groups.c.domain_name == params["domain"])
                .where(groups.c.name == group_id_or_name)
            )
            group_id = await conn.scalar(query)
            domain = params["domain"]
        else:
            group_id = group_id_or_name
            domain = await query_group_domain(conn, group_id)
        if group_id is None or domain is None:
            raise ProjectNotFound
        if not request["is_superadmin"]:
            if request["is_admin"]:
                if request["user"]["domain_name"] != domain:
                    raise GenericForbidden(
                        "Domain admins cannot access group dotfiles of other domains"
                    )
            else:
                # check if user (non-admin) is in the group
                query = (
                    sa.select(agus.c.group_id)
                    .select_from(agus)
                    .where(agus.c.user_id == request["user"]["uuid"])
                )
                result = await conn.execute(query)
                rows = result.fetchall()
                if group_id not in map(lambda x: x.group_id, rows):
                    raise GenericForbidden("Users cannot access group dotfiles of other groups")

        if params["path"]:
            dotfiles, _ = await query_group_dotfiles(conn, group_id)
            if dotfiles is None:
                raise ProjectNotFound
            for dotfile in dotfiles:
                if dotfile["path"] == params["path"]:
                    return web.json_response(dotfile)
            raise DotfileNotFound
        dotfiles, _ = await query_group_dotfiles(conn, group_id)
        if dotfiles is None:
            raise ProjectNotFound
        for entry in dotfiles:
            resp.append({
                "path": entry["path"],
                "permission": entry["perm"],
                "data": entry["data"],
            })
        return web.json_response(resp)


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict(
        {
            tx.AliasedKey(["group", "groupId", "group_id"]): tx.UUID | t.String,
            t.Key("domain", default=None): t.String | t.Null,
            t.Key("data"): t.String(max_length=MAXIMUM_DOTFILE_SIZE),
            t.Key("path"): t.String,
            t.Key("permission"): t.Regexp(r"^[0-7]{3}$", re.ASCII),
        },
    )
)
async def update(request: web.Request, params: Any) -> web.Response:
    log.info("GROUPCONFIG.UPDATE_DOTFILE (domain:{0})", params["domain"])
    root_ctx: RootContext = request.app["_root.context"]
    group_id_or_name = params["group"]
    async with root_ctx.db.begin() as conn:
        if isinstance(group_id_or_name, str):
            if params["domain"] is None:
                raise InvalidAPIParameters("Missing parameter 'domain'")
            query = (
                sa.select(groups.c.id)
                .select_from(groups)
                .where(groups.c.domain_name == params["domain"])
                .where(groups.c.name == group_id_or_name)
            )
            group_id = await conn.scalar(query)
            domain = params["domain"]
        else:
            group_id = group_id_or_name
            domain = await query_group_domain(conn, group_id)
        if group_id is None or domain is None:
            raise ProjectNotFound
        if not request["is_superadmin"] and request["user"]["domain_name"] != domain:
            raise GenericForbidden("Admins cannot update group dotfiles of other domains")

        dotfiles, _ = await query_group_dotfiles(conn, group_id)
        if dotfiles is None:
            raise ProjectNotFound
        new_dotfiles = [x for x in dotfiles if x["path"] != params["path"]]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound

        new_dotfiles.append({
            "path": params["path"],
            "perm": params["permission"],
            "data": params["data"],
        })
        dotfile_packed = msgpack.packb(new_dotfiles)
        if len(dotfile_packed) > MAXIMUM_DOTFILE_SIZE:
            raise DotfileCreationFailed("No leftover space for dotfile storage")

        query = groups.update().values(dotfiles=dotfile_packed).where(groups.c.id == group_id)
        await conn.execute(query)
    return web.json_response({})


@server_status_required(READ_ALLOWED)
@admin_required
@check_api_params(
    t.Dict({
        tx.AliasedKey(["group", "groupId", "group_id"]): tx.UUID | t.String,
        t.Key("domain", default=None): t.String | t.Null,
        t.Key("path"): t.String,
    }),
)
async def delete(request: web.Request, params: Any) -> web.Response:
    log.info("GROUPCONFIG.DELETE_DOTFILE (domain:{0})", params["domain"])
    root_ctx: RootContext = request.app["_root.context"]
    group_id_or_name = params["group"]
    async with root_ctx.db.begin() as conn:
        if isinstance(group_id_or_name, str):
            if params["domain"] is None:
                raise InvalidAPIParameters("Missing parameter 'domain'")
            query = (
                sa.select(groups.c.id)
                .select_from(groups)
                .where(groups.c.domain_name == params["domain"])
                .where(groups.c.name == group_id_or_name)
            )
            group_id = await conn.scalar(query)
            domain = params["domain"]
        else:
            group_id = group_id_or_name
            domain = await query_group_domain(conn, group_id)
        if group_id is None or domain is None:
            raise ProjectNotFound
        if not request["is_superadmin"] and request["user"]["domain_name"] != domain:
            raise GenericForbidden("Admins cannot delete dotfiles of other domains")

        dotfiles, _ = await query_group_dotfiles(conn, group_id)
        if dotfiles is None:
            raise DotfileNotFound
        new_dotfiles = [x for x in dotfiles if x["path"] != params["path"]]
        if len(new_dotfiles) == len(dotfiles):
            raise DotfileNotFound

        dotfile_packed = msgpack.packb(new_dotfiles)
        query = groups.update().values(dotfiles=dotfile_packed).where(groups.c.id == group_id)
        await conn.execute(query)
        return web.json_response({"success": True})


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["api_versions"] = (4, 5)
    app["prefix"] = "group-config"
    cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    cors.add(app.router.add_route("POST", "/dotfiles", create))
    cors.add(app.router.add_route("GET", "/dotfiles", list_or_get))
    cors.add(app.router.add_route("PATCH", "/dotfiles", update))
    cors.add(app.router.add_route("DELETE", "/dotfiles", delete))

    return app, []



"""migrate project data to roles

Revision ID: 430b1631804d
Revises: a4d56e86d9ee
Create Date: 2025-08-12 15:30:10.369554

"""

import enum
import uuid
from collections.abc import Iterable, Mapping, Sequence
from typing import Any

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects.postgresql import UUID as PGUUID
from sqlalchemy.engine import Connection, Row

from ai.backend.manager.data.permission.id import ObjectId, ScopeId
from ai.backend.manager.models.base import GUID, EnumValueType, IDColumn
from ai.backend.manager.models.rbac_models.migration.enums import (
    EntityType,
    OperationType,
    RoleSource,
    ScopeType,
)
from ai.backend.manager.models.rbac_models.migration.models import (
    get_association_scopes_entities_table,
    get_permission_groups_table,
    get_permissions_table,
    get_roles_table,
    get_user_roles_table,
    mapper_registry,
)
from ai.backend.manager.models.rbac_models.migration.project import (
    ADMIN_ACCESSIBLE_ENTITY_TYPES_IN_PROJECT,
    MEMBER_ACCESSIBLE_ENTITY_TYPES_IN_PROJECT,
    ProjectData,
    get_project_admin_role_creation_input,
    get_project_member_role_creation_input,
)
from ai.backend.manager.models.rbac_models.migration.types import (
    AssociationScopesEntitiesCreateInput,
    PermissionCreateInput,
    PermissionGroupCreateInput,
    UserRoleCreateInput,
)
from ai.backend.manager.models.rbac_models.migration.utils import (
    insert_if_data_exists,
    query_permission_groups_by_scope_ids,
    query_role_rows_by_name,
)

# revision identifiers, used by Alembic.
revision = "430b1631804d"
down_revision = "a4d56e86d9ee"
branch_labels = None
depends_on = None


class UserRole(enum.StrEnum):
    """
    User's role.
    """

    SUPERADMIN = "superadmin"
    ADMIN = "admin"
    USER = "user"
    MONITOR = "monitor"


class Tables:
    @staticmethod
    def get_users_table() -> sa.Table:
        return sa.Table(
            "users",
            mapper_registry.metadata,
            IDColumn("uuid"),
            sa.Column("username", sa.String(length=64), unique=True),
            sa.Column("domain_name", sa.String(length=64), index=True),
            sa.Column("role", EnumValueType(UserRole), default=UserRole.USER),
            extend_existing=True,
        )

    @staticmethod
    def get_groups_table() -> sa.Table:
        return sa.Table(
            "groups",
            mapper_registry.metadata,
            IDColumn(),
            sa.Column("name", sa.String(64), nullable=False, unique=True),
            sa.Column(
                "domain_name",
                sa.String(length=64),
                sa.ForeignKey("domains.name", onupdate="CASCADE", ondelete="CASCADE"),
                nullable=False,
                index=True,
            ),
            extend_existing=True,
        )

    @staticmethod
    def get_association_groups_users_table() -> sa.Table:
        return sa.Table(
            "association_groups_users",
            mapper_registry.metadata,
            IDColumn(),
            sa.Column("user_id", GUID, nullable=False),
            sa.Column(
                "group_id",
                GUID,
                sa.ForeignKey("groups.id", onupdate="CASCADE", ondelete="CASCADE"),
                nullable=False,
            ),
            extend_existing=True,
        )


class RoleCreator:
    @classmethod
    def _create_admin_roles(
        cls, db_conn: Connection, rows: Sequence[Row[Any]]
    ) -> dict[str, uuid.UUID]:
        roles_table = get_roles_table()
        role_inputs: list[dict[str, Any]] = []
        role_name_project_id_map: dict[str, uuid.UUID] = {}
        for row in rows:
            data = ProjectData.from_row(row)
            admin_role_input = get_project_admin_role_creation_input(data)
            role_inputs.append(admin_role_input.to_dict())
            role_name_project_id_map[admin_role_input.name] = data.id

        insert_if_data_exists(db_conn, roles_table, role_inputs)
        return role_name_project_id_map

    @classmethod
    def _create_member_roles(
        cls, db_conn: Connection, rows: Sequence[Row[Any]]
    ) -> dict[str, uuid.UUID]:
        roles_table = get_roles_table()
        role_inputs: list[dict[str, Any]] = []
        role_name_project_id_map: dict[str, uuid.UUID] = {}
        for row in rows:
            data = ProjectData.from_row(row)
            member_role_input = get_project_member_role_creation_input(data)
            role_inputs.append(member_role_input.to_dict())
            role_name_project_id_map[member_role_input.name] = data.id

        insert_if_data_exists(db_conn, roles_table, role_inputs)
        return role_name_project_id_map

    @classmethod
    def _create_permission_groups(
        cls, db_conn: Connection, role_id_project_id_map: Mapping[uuid.UUID, uuid.UUID]
    ) -> None:
        permission_groups_table = get_permission_groups_table()
        permission_group_inputs: list[dict[str, Any]] = []
        for role_id, project_id in role_id_project_id_map.items():
            input = PermissionGroupCreateInput(
                role_id=role_id,
                scope_type=ScopeType.PROJECT,
                scope_id=str(project_id),
            )
            permission_group_inputs.append(input.to_dict())
        insert_if_data_exists(db_conn, permission_groups_table, permission_group_inputs)

    @classmethod
    def _create_admin_permissions(
        cls, db_conn: Connection, permission_group_ids: Iterable[uuid.UUID]
    ) -> None:
        scope_permission_inputs: list[dict[str, Any]] = []
        for permission_group_id in permission_group_ids:
            for entity_type in ADMIN_ACCESSIBLE_ENTITY_TYPES_IN_PROJECT:
                for operation in OperationType.admin_operations():
                    input = PermissionCreateInput(
                        permission_group_id=permission_group_id,
                        entity_type=entity_type,
                        operation=operation,
                    )
                    scope_permission_inputs.append(input.to_dict())
        insert_if_data_exists(db_conn, get_permissions_table(), scope_permission_inputs)

    @classmethod
    def _create_member_permissions(
        cls, db_conn: Connection, permission_group_ids: Iterable[uuid.UUID]
    ) -> None:
        scope_permission_inputs: list[dict[str, Any]] = []
        for permission_group_id in permission_group_ids:
            for entity_type in MEMBER_ACCESSIBLE_ENTITY_TYPES_IN_PROJECT:
                for operation in OperationType.member_operations():
                    input = PermissionCreateInput(
                        permission_group_id=permission_group_id,
                        entity_type=entity_type,
                        operation=operation,
                    )
                    scope_permission_inputs.append(input.to_dict())
        insert_if_data_exists(db_conn, get_permissions_table(), scope_permission_inputs)

    @classmethod
    def _create_project_admin_roles_and_permissions(
        cls, db_conn: Connection, rows: Sequence[Row[Any]]
    ) -> None:
        role_name_project_id_map = cls._create_admin_roles(db_conn, rows)
        role_rows = query_role_rows_by_name(db_conn, list(role_name_project_id_map.keys()))
        role_id_project_id_map: dict[uuid.UUID, uuid.UUID] = {
            row.id: role_name_project_id_map[row.name] for row in role_rows
        }
        cls._create_permission_groups(db_conn, role_id_project_id_map)
        str_project_ids = [str(user_id) for user_id in role_id_project_id_map.values()]
        permission_group_ids = query_permission_groups_by_scope_ids(db_conn, str_project_ids)
        cls._create_admin_permissions(db_conn, permission_group_ids)

    @classmethod
    def _create_project_member_roles_and_permissions(
        cls, db_conn: Connection, rows: Sequence[Row[Any]]
    ) -> None:
        role_name_project_id_map = cls._create_member_roles(db_conn, rows)
        role_rows = query_role_rows_by_name(db_conn, list(role_name_project_id_map.keys()))
        role_id_project_id_map: dict[uuid.UUID, uuid.UUID] = {
            row.id: role_name_project_id_map[row.name] for row in role_rows
        }
        cls._create_permission_groups(db_conn, role_id_project_id_map)
        str_project_ids = [str(user_id) for user_id in role_id_project_id_map.values()]
        permission_group_ids = query_permission_groups_by_scope_ids(db_conn, str_project_ids)
        cls._create_member_permissions(db_conn, permission_group_ids)

    @classmethod
    def _query_project_row(
        cls, db_conn: Connection, offset: int, page_size: int
    ) -> Sequence[Row[Any]]:
        """
        Query all project rows with pagination.
        """
        groups_table = Tables.get_groups_table()
        project_query = (
            sa.select(groups_table.c.id, groups_table.c.domain_name)
            .offset(offset)
            .limit(page_size)
            .order_by(groups_table.c.id)
        )
        return db_conn.execute(project_query).all()

    @classmethod
    def create_project_roles_and_permissions(cls, db_conn: Connection) -> None:
        """
        Migrate project data to roles and permissions.
        All projects have a default admin role and a member role.
        """
        offset = 0
        page_size = 1000
        while True:
            rows = cls._query_project_row(db_conn, offset, page_size)
            offset += page_size
            if not rows:
                break
            cls._create_project_admin_roles_and_permissions(db_conn, rows)
            cls._create_project_member_roles_and_permissions(db_conn, rows)


class RoleMapper:
    @classmethod
    def _query_admin_role_with_user(
        cls, db_conn: Connection, offset: int, page_size: int
    ) -> Sequence[Row[Any]]:
        """
        Query admin role rows with their associated users.
        All projects in a domain is queried with domain admin users.
        """
        users_table = Tables.get_users_table()
        roles_table = get_roles_table()
        groups_table = Tables.get_groups_table()

        permission_groups_table = get_permission_groups_table()
        stmt = (
            sa.select(
                roles_table.c.id.label("role_id"),
                permission_groups_table.c.scope_id,
                users_table.c.uuid.label("user_id"),
            )
            .select_from(
                sa.join(
                    roles_table,
                    permission_groups_table,
                    roles_table.c.id == permission_groups_table.c.role_id,
                )
                .join(
                    groups_table,
                    sa.cast(permission_groups_table.c.scope_id, PGUUID) == groups_table.c.id,
                )
                .join(
                    users_table,
                    groups_table.c.domain_name == users_table.c.domain_name,
                )
            )
            .where(
                sa.and_(
                    roles_table.c.source == RoleSource.SYSTEM,
                    users_table.c.role == UserRole.ADMIN,
                )
            )
            .offset(offset)
            .limit(page_size)
            .order_by(users_table.c.uuid)
        )
        return db_conn.execute(stmt).all()

    @classmethod
    def _map_admin_users_to_project(cls, db_conn: Connection) -> None:
        user_roles_table = get_user_roles_table()
        association_table = get_association_scopes_entities_table()

        offset = 0
        page_size = 1000

        while True:
            rows = cls._query_admin_role_with_user(db_conn, offset, page_size)
            offset += page_size
            if not rows:
                break
            user_role_inputs: list[dict[str, Any]] = []
            association_inputs: list[dict[str, Any]] = []
            for row in rows:
                input = UserRoleCreateInput(user_id=row.user_id, role_id=row.role_id)
                user_role_inputs.append(input.to_dict())

                association_input = AssociationScopesEntitiesCreateInput(
                    scope_id=ScopeId(
                        scope_type=ScopeType.PROJECT.to_original(),
                        scope_id=row.scope_id,
                    ),
                    object_id=ObjectId(
                        entity_type=EntityType.USER.to_original(),
                        entity_id=str(row.user_id),
                    ),
                )
                association_inputs.append(association_input.to_dict())
            insert_if_data_exists(db_conn, user_roles_table, user_role_inputs)
            insert_if_data_exists(db_conn, association_table, association_inputs)

    @classmethod
    def _query_member_role_with_user(
        cls, db_conn: Connection, offset: int, page_size: int
    ) -> Sequence[Row[Any]]:
        """
        Query member role rows with their associated users.
        """
        users_table = Tables.get_users_table()
        roles_table = get_roles_table()
        association_groups_users_table = Tables.get_association_groups_users_table()

        permission_groups_table = get_permission_groups_table()
        stmt = (
            sa.select(
                roles_table.c.id.label("role_id"),
                permission_groups_table.c.scope_id,
                users_table.c.uuid.label("user_id"),
            )
            .select_from(
                sa.join(
                    roles_table,
                    permission_groups_table,
                    roles_table.c.id == permission_groups_table.c.role_id,
                )
                .join(
                    association_groups_users_table,
                    sa.cast(permission_groups_table.c.scope_id, PGUUID)
                    == association_groups_users_table.c.group_id,
                )
                .join(
                    users_table,
                    association_groups_users_table.c.user_id == users_table.c.uuid,
                )
            )
            .where(
                sa.and_(
                    roles_table.c.source == RoleSource.CUSTOM,
                    users_table.c.role == UserRole.USER,
                )
            )
            .offset(offset)
            .limit(page_size)
            .order_by(users_table.c.uuid)
        )
        return db_conn.execute(stmt).all()

    @classmethod
    def _map_member_users_to_project(cls, db_conn: Connection) -> None:
        user_roles_table = get_user_roles_table()
        association_table = get_association_scopes_entities_table()

        offset = 0
        page_size = 1000

        while True:
            rows = cls._query_member_role_with_user(db_conn, offset, page_size)
            offset += page_size
            if not rows:
                break
            user_role_inputs: list[dict[str, Any]] = []
            association_inputs: list[dict[str, Any]] = []
            for row in rows:
                input = UserRoleCreateInput(user_id=row.user_id, role_id=row.role_id)
                user_role_inputs.append(input.to_dict())

                association_input = AssociationScopesEntitiesCreateInput(
                    scope_id=ScopeId(
                        scope_type=ScopeType.PROJECT.to_original(),
                        scope_id=row.scope_id,
                    ),
                    object_id=ObjectId(
                        entity_type=EntityType.USER.to_original(),
                        entity_id=str(row.user_id),
                    ),
                )
                association_inputs.append(association_input.to_dict())
            insert_if_data_exists(db_conn, user_roles_table, user_role_inputs)
            insert_if_data_exists(db_conn, association_table, association_inputs)

    @classmethod
    def map_users_to_project(cls, db_conn: Connection) -> None:
        cls._map_admin_users_to_project(db_conn)
        cls._map_member_users_to_project(db_conn)


def upgrade() -> None:
    op.execute("""
        ALTER TABLE association_scopes_entities
        DROP CONSTRAINT IF EXISTS uq_scope_id_entity_id
    """)
    conn = op.get_bind()
    RoleCreator.create_project_roles_and_permissions(conn)
    RoleMapper.map_users_to_project(conn)


def downgrade() -> None:
    pass



"""Central registry for notification channel handlers."""

from __future__ import annotations

import logging
from datetime import datetime

import jinja2

from ai.backend.common.clients.http_client import ClientPool, tcp_client_session_factory
from ai.backend.common.data.notification import (
    NotifiableMessage,
    NotificationChannelType,
)
from ai.backend.common.data.notification.types import EmailSpec, WebhookSpec
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.notification import NotificationChannelData
from ai.backend.manager.errors.notification import (
    InvalidNotificationChannelType,
    NotificationTemplateRenderingFailure,
)
from ai.backend.manager.notification.channels.email.channel import EmailChannel
from ai.backend.manager.notification.channels.webhook.channel import WebhookChannel

from .channels.base import AbstractNotificationChannel
from .types import NotificationMessage, ProcessRuleParams, SendResult

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("NotificationCenter",)


class NotificationCenter:
    """
    Central registry for notification channel handlers.

    This class manages the available notification channel handlers and
    provides a unified interface for sending notifications through different channels.
    """

    _http_client_pool: ClientPool
    _template_env: jinja2.Environment

    def __init__(self) -> None:
        """Initialize the notification center with HTTP client pool and template environment."""
        self._http_client_pool = ClientPool(
            factory=tcp_client_session_factory,
            cleanup_interval_seconds=600,
        )
        self._template_env = jinja2.Environment(
            loader=jinja2.BaseLoader(),
            autoescape=False,
        )

    async def close(self) -> None:
        """Close the HTTP client pool and cleanup resources."""
        await self._http_client_pool.close()

    def _create_handler(
        self,
        channel_data: NotificationChannelData,
    ) -> AbstractNotificationChannel:
        """
        Create a handler for the given channel.

        Args:
            channel_data: Channel configuration and metadata

        Returns:
            Handler instance for the channel
        """
        # Create handler based on channel type (no caching due to dynamic config changes)
        match channel_data.channel_type:
            case NotificationChannelType.WEBHOOK:
                if not isinstance(channel_data.spec, WebhookSpec):
                    raise InvalidNotificationChannelType(
                        f"Invalid config type for WEBHOOK channel: {type(channel_data.spec)}"
                    )
                return WebhookChannel(
                    http_client_pool=self._http_client_pool,
                    webhook_spec=channel_data.spec,
                )

            case NotificationChannelType.EMAIL:
                if not isinstance(channel_data.spec, EmailSpec):
                    raise InvalidNotificationChannelType(
                        f"Invalid config type for EMAIL channel: {type(channel_data.spec)}"
                    )
                return EmailChannel(email_spec=channel_data.spec)

        # Explicitly handle unregistered channel types
        raise InvalidNotificationChannelType(
            f"No handler registered for channel type: {channel_data.channel_type}"
        )

    async def process_rule(
        self,
        params: ProcessRuleParams,
    ) -> SendResult:
        """
        Process notification rule: render template and send notification.

        Args:
            params: Parameters for rule processing

        Returns:
            SendResult indicating success

        Raises:
            Exception: If template rendering or notification delivery fails
        """
        # Render message template
        rendered_message = self._render_template(
            params.message_template,
            params.rule_type,
            params.timestamp,
            params.notification_data,
        )

        # Create notification message
        message = NotificationMessage(
            message=rendered_message,
        )

        # Send notification through channel
        handler = self._create_handler(params.channel)
        return await handler.send(message)

    async def validate_channel(
        self,
        channel_data: NotificationChannelData,
        test_message: str,
    ) -> SendResult:
        """
        Validate a notification channel by sending a test message.

        Args:
            channel_data: Channel configuration and metadata
            test_message: Test message to send

        Returns:
            SendResult indicating success

        Raises:
            NotificationProcessingFailure: If validation fails
        """
        # Create notification message from provided test message
        message = NotificationMessage(
            message=test_message,
        )

        # Send the test message
        handler = self._create_handler(channel_data)
        return await handler.send(message)

    def _render_template(
        self,
        template_str: str,
        rule_type: str,
        timestamp: datetime,
        notification_data: NotifiableMessage,
    ) -> str:
        """Render Jinja2 template with notification event data."""
        try:
            # Convert Pydantic model to dict for template context
            data_dict = notification_data.model_dump()
            template = self._template_env.from_string(template_str)
            return template.render(
                rule_type=str(rule_type),
                timestamp=timestamp,
                data=data_dict,
                **data_dict,  # Allow direct access to data fields
            )
        except jinja2.TemplateError as e:
            log.error(
                "Failed to render notification template",
                template=template_str,
                error=str(e),
            )
            raise NotificationTemplateRenderingFailure(f"Failed to render template: {e!s}") from e



"""Export DB source for database operations."""

from __future__ import annotations

from collections.abc import AsyncIterator, Sequence
from typing import TYPE_CHECKING, Any

from ai.backend.manager.repositories.base.export import (
    StreamingExportQuery,
    execute_streaming_export,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.utils import ExtendedAsyncSAEngine

__all__ = ("ExportDBSource",)


class ExportDBSource:
    """DB source for export streaming operations."""

    _db: ExtendedAsyncSAEngine

    def __init__(self, db: ExtendedAsyncSAEngine) -> None:
        self._db = db

    async def stream_export(
        self,
        query: StreamingExportQuery,
    ) -> AsyncIterator[Sequence[Sequence[Any]]]:
        """Execute streaming export.

        Args:
            query: Export query containing fields, conditions, orders, and limits

        Yields:
            Partitions of row values (each row in query.fields order)
        """
        async for partition in execute_streaming_export(self._db, query):
            yield partition



"""UpdaterSpec implementations for deployment repository."""

from __future__ import annotations

from dataclasses import dataclass, field
from decimal import Decimal
from typing import Any, override
from uuid import UUID

from ai.backend.common.data.model_deployment.types import DeploymentStrategy
from ai.backend.common.types import AutoScalingMetricComparator, AutoScalingMetricSource
from ai.backend.manager.data.deployment.types import RouteStatus, RouteTrafficStatus
from ai.backend.manager.models.deployment_auto_scaling_policy import (
    DeploymentAutoScalingPolicyRow,
)
from ai.backend.manager.models.deployment_policy import (
    BlueGreenSpec,
    DeploymentPolicyRow,
    RollingUpdateSpec,
)
from ai.backend.manager.models.endpoint import EndpointRow
from ai.backend.manager.models.routing import RoutingRow
from ai.backend.manager.repositories.base.updater import UpdaterSpec
from ai.backend.manager.types import OptionalState, TriState


@dataclass
class DeploymentMetadataUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for deployment metadata updates."""

    name: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    domain: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    project: OptionalState[UUID] = field(default_factory=OptionalState[UUID].nop)
    resource_group: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    tag: TriState[str] = field(default_factory=TriState[str].nop)

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.name.update_dict(to_update, "name")
        self.domain.update_dict(to_update, "domain")
        self.project.update_dict(to_update, "project")
        self.resource_group.update_dict(to_update, "resource_group")
        self.tag.update_dict(to_update, "tag")
        return to_update


@dataclass
class ReplicaSpecUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for replica specification updates."""

    replica_count: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    desired_replica_count: OptionalState[int] = field(default_factory=OptionalState[int].nop)

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        # Use the actual database column names
        self.replica_count.update_dict(to_update, "replicas")
        self.desired_replica_count.update_dict(to_update, "desired_replicas")
        return to_update


@dataclass
class DeploymentNetworkSpecUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for deployment network specification updates."""

    open_to_public: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)
    url: TriState[str] = field(default_factory=TriState[str].nop)

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.open_to_public.update_dict(to_update, "open_to_public")
        self.url.update_dict(to_update, "url")
        return to_update


@dataclass
class MountUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for mount-related updates."""

    model_vfolder_id: OptionalState[UUID] = field(default_factory=OptionalState[UUID].nop)

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.model_vfolder_id.update_dict(to_update, "model_vfolder_id")
        return to_update


@dataclass
class RevisionStateUpdaterSpec(UpdaterSpec[EndpointRow]):
    """UpdaterSpec for deployment revision state updates.

    Manages which revision is currently active and which is being deployed.
    """

    current_revision: TriState[UUID] = field(default_factory=TriState[UUID].nop)
    deploying_revision: TriState[UUID] = field(default_factory=TriState[UUID].nop)
    revision_history_limit: OptionalState[int] = field(default_factory=OptionalState[int].nop)

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.current_revision.update_dict(to_update, "current_revision")
        self.deploying_revision.update_dict(to_update, "deploying_revision")
        self.revision_history_limit.update_dict(to_update, "revision_history_limit")
        return to_update


@dataclass
class DeploymentUpdaterSpec(UpdaterSpec[EndpointRow]):
    """Composite UpdaterSpec for deployment updates.

    Combines metadata, replica_spec, network, mount, and revision_state updates.
    """

    metadata: DeploymentMetadataUpdaterSpec | None = None
    replica_spec: ReplicaSpecUpdaterSpec | None = None
    network: DeploymentNetworkSpecUpdaterSpec | None = None
    mount: MountUpdaterSpec | None = None
    revision_state: RevisionStateUpdaterSpec | None = None

    @property
    @override
    def row_class(self) -> type[EndpointRow]:
        return EndpointRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        if self.metadata:
            to_update.update(self.metadata.build_values())
        if self.replica_spec:
            to_update.update(self.replica_spec.build_values())
        if self.network:
            to_update.update(self.network.build_values())
        if self.mount:
            to_update.update(self.mount.build_values())
        if self.revision_state:
            to_update.update(self.revision_state.build_values())
        return to_update


@dataclass
class DeploymentAutoScalingPolicyUpdaterSpec(UpdaterSpec[DeploymentAutoScalingPolicyRow]):
    """UpdaterSpec for deployment auto-scaling policy updates.

    All fields are optional - only specified fields will be updated.
    Supports partial updates for hysteresis-based scaling configuration.
    """

    min_replicas: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    max_replicas: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    metric_source: TriState[AutoScalingMetricSource] = field(
        default_factory=TriState[AutoScalingMetricSource].nop
    )
    metric_name: TriState[str] = field(default_factory=TriState[str].nop)
    comparator: TriState[AutoScalingMetricComparator] = field(
        default_factory=TriState[AutoScalingMetricComparator].nop
    )
    scale_up_threshold: TriState[Decimal] = field(default_factory=TriState[Decimal].nop)
    scale_down_threshold: TriState[Decimal] = field(default_factory=TriState[Decimal].nop)
    scale_up_step_size: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    scale_down_step_size: OptionalState[int] = field(default_factory=OptionalState[int].nop)
    cooldown_seconds: OptionalState[int] = field(default_factory=OptionalState[int].nop)

    @property
    @override
    def row_class(self) -> type[DeploymentAutoScalingPolicyRow]:
        return DeploymentAutoScalingPolicyRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.min_replicas.update_dict(to_update, "min_replicas")
        self.max_replicas.update_dict(to_update, "max_replicas")
        self.metric_source.update_dict(to_update, "metric_source")
        self.metric_name.update_dict(to_update, "metric_name")
        self.comparator.update_dict(to_update, "comparator")
        self.scale_up_threshold.update_dict(to_update, "scale_up_threshold")
        self.scale_down_threshold.update_dict(to_update, "scale_down_threshold")
        self.scale_up_step_size.update_dict(to_update, "scale_up_step_size")
        self.scale_down_step_size.update_dict(to_update, "scale_down_step_size")
        self.cooldown_seconds.update_dict(to_update, "cooldown_seconds")
        return to_update


@dataclass
class DeploymentPolicyUpdaterSpec(UpdaterSpec[DeploymentPolicyRow]):
    """UpdaterSpec for deployment policy updates.

    All fields are optional - only specified fields will be updated.
    Supports partial updates for deployment strategy configuration.
    """

    strategy: OptionalState[DeploymentStrategy] = field(
        default_factory=OptionalState[DeploymentStrategy].nop
    )
    strategy_spec: OptionalState[RollingUpdateSpec | BlueGreenSpec] = field(
        default_factory=OptionalState[RollingUpdateSpec | BlueGreenSpec].nop
    )
    rollback_on_failure: OptionalState[bool] = field(default_factory=OptionalState[bool].nop)

    @property
    @override
    def row_class(self) -> type[DeploymentPolicyRow]:
        return DeploymentPolicyRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.strategy.update_dict(to_update, "strategy")
        # Convert Pydantic model to dict for JSONB storage
        spec_value = self.strategy_spec.optional_value()
        if spec_value is not None:
            to_update["strategy_spec"] = spec_value.model_dump()
        self.rollback_on_failure.update_dict(to_update, "rollback_on_failure")
        return to_update


@dataclass
class RouteStatusUpdaterSpec(UpdaterSpec[RoutingRow]):
    """UpdaterSpec for route status updates.

    Updates health status and traffic status of a route.
    """

    status: OptionalState[RouteStatus] = field(default_factory=OptionalState[RouteStatus].nop)
    traffic_status: OptionalState[RouteTrafficStatus] = field(
        default_factory=OptionalState[RouteTrafficStatus].nop
    )

    @property
    @override
    def row_class(self) -> type[RoutingRow]:
        return RoutingRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.status.update_dict(to_update, "status")
        self.traffic_status.update_dict(to_update, "traffic_status")
        return to_update


@dataclass
class RouteSessionUpdaterSpec(UpdaterSpec[RoutingRow]):
    """UpdaterSpec for route session binding.

    Binds a session to a route for traffic routing.
    """

    session: UUID

    @property
    @override
    def row_class(self) -> type[RoutingRow]:
        return RoutingRow

    @override
    def build_values(self) -> dict[str, Any]:
        return {
            "session": self.session,
            "status": RouteStatus.PROVISIONING,
        }


@dataclass
class RouteUpdaterSpec(UpdaterSpec[RoutingRow]):
    """Unified UpdaterSpec for route updates.

    Combines all route update operations into a single spec.
    Each field uses OptionalState to support partial updates.
    """

    status: OptionalState[RouteStatus] = field(default_factory=OptionalState[RouteStatus].nop)
    traffic_status: OptionalState[RouteTrafficStatus] = field(
        default_factory=OptionalState[RouteTrafficStatus].nop
    )
    session: OptionalState[UUID] = field(default_factory=OptionalState[UUID].nop)
    traffic_ratio: OptionalState[float] = field(default_factory=OptionalState[float].nop)
    revision: OptionalState[UUID] = field(default_factory=OptionalState[UUID].nop)
    error_data: OptionalState[dict[str, Any]] = field(
        default_factory=OptionalState[dict[str, Any]].nop
    )

    @property
    @override
    def row_class(self) -> type[RoutingRow]:
        return RoutingRow

    @override
    def build_values(self) -> dict[str, Any]:
        to_update: dict[str, Any] = {}
        self.status.update_dict(to_update, "status")
        self.traffic_status.update_dict(to_update, "traffic_status")
        self.session.update_dict(to_update, "session")
        self.traffic_ratio.update_dict(to_update, "traffic_ratio")
        self.revision.update_dict(to_update, "revision")
        self.error_data.update_dict(to_update, "error_data")
        return to_update



from collections.abc import Iterable

# import aiohttp_cors
from aiohttp import web

from .types import CORSOptions, WebMiddleware


async def init(app: web.Application) -> None:
    pass


async def shutdown(app: web.Application) -> None:
    pass


def create_app(
    _default_cors_options: CORSOptions,
) -> tuple[web.Application, Iterable[WebMiddleware]]:
    app = web.Application()
    app.on_startup.append(init)
    app.on_shutdown.append(shutdown)
    app["prefix"] = "image"
    app["api_versions"] = (4,)
    # cors = aiohttp_cors.setup(app, defaults=default_cors_options)
    return app, []



"""Add artifact_registries table

Revision ID: ba5207c55924
Revises: 11ee5c074009
Create Date: 2025-08-26 06:29:20.265879

"""

import sqlalchemy as sa
from alembic import op

from ai.backend.manager.models.base import GUID

# revision identifiers, used by Alembic.
revision = "ba5207c55924"
down_revision = "11ee5c074009"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table(
        "artifact_registries",
        sa.Column("id", GUID(), server_default=sa.text("uuid_generate_v4()"), nullable=False),
        sa.Column("name", sa.String(), nullable=False),
        sa.Column("registry_id", GUID(), nullable=False),
        sa.Column("type", sa.String(), nullable=False),
        sa.PrimaryKeyConstraint("id", name=op.f("pk_artifact_registries")),
        sa.UniqueConstraint("name", name=op.f("uq_artifact_registries_name")),
        sa.UniqueConstraint("registry_id", name=op.f("uq_artifact_registries_registry_id")),
    )

    conn = op.get_bind()
    inspector = sa.inspect(conn)

    # Check if huggingface_registries table exists
    if inspector.has_table("huggingface_registries"):
        columns = [col["name"] for col in inspector.get_columns("huggingface_registries")]

        # Migrate existing huggingface_registries records to artifact_registries
        result = conn.execute(
            sa.text("""
            SELECT id, name FROM huggingface_registries
            WHERE name IS NOT NULL
        """)
        )

        for row in result:
            conn.execute(
                sa.text("""
                INSERT INTO artifact_registries (name, registry_id, type)
                VALUES (:name, :registry_id, :type)
            """),
                {"name": row.name, "registry_id": row.id, "type": "huggingface"},
            )

        # Drop the name column and constraint if they exist
        if "name" in columns:
            op.drop_constraint(
                "uq_huggingface_registries_name", "huggingface_registries", type_="unique"
            )
            op.drop_column("huggingface_registries", "name")

    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    inspector = sa.inspect(conn)

    # Add name column as nullable first to avoid constraint violations
    op.add_column(
        "huggingface_registries",
        sa.Column("name", sa.VARCHAR(), autoincrement=False, nullable=True),
    )

    # Check if artifact_registries table exists and migrate data back
    if inspector.has_table("artifact_registries"):
        # Migrate data back from artifact_registries to huggingface_registries
        result = conn.execute(
            sa.text("""
            SELECT name, registry_id FROM artifact_registries
            WHERE type = 'huggingface'
        """)
        )

        for row in result:
            conn.execute(
                sa.text("""
                UPDATE huggingface_registries
                SET name = :name
                WHERE id = :registry_id
            """),
                {"name": row.name, "registry_id": row.registry_id},
            )

    # Now make the name column NOT NULL after populating data
    op.alter_column(
        "huggingface_registries",
        "name",
        existing_type=sa.VARCHAR(),
        nullable=False,
    )

    # Create the unique constraint after ensuring no null values
    op.create_unique_constraint(
        "uq_huggingface_registries_name", "huggingface_registries", ["name"]
    )
    op.drop_table("artifact_registries")
    # ### end Alembic commands ###



from dataclasses import dataclass, field
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.group.types import GroupData
from ai.backend.manager.models.group import GroupRow
from ai.backend.manager.repositories.base.updater import Updater
from ai.backend.manager.services.group.actions.base import GroupAction
from ai.backend.manager.types import OptionalState


@dataclass
class ModifyGroupAction(GroupAction):
    updater: Updater[GroupRow]
    user_update_mode: OptionalState[str] = field(default_factory=OptionalState[str].nop)
    user_uuids: OptionalState[list[str]] = field(default_factory=OptionalState[list[str]].nop)

    @override
    def entity_id(self) -> str | None:
        return str(self.updater.pk_value)

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.UPDATE

    def update_mode(self) -> str | None:
        if self.user_uuids.optional_value():
            return self.user_update_mode.optional_value()
        return None


@dataclass
class ModifyGroupActionResult(BaseActionResult):
    data: GroupData | None

    @override
    def entity_id(self) -> str | None:
        return str(self.data.id) if self.data else None



from __future__ import annotations

import logging
import uuid
from typing import TYPE_CHECKING

import sqlalchemy as sa
from sqlalchemy.orm import Mapped, foreign, mapped_column, relationship

from ai.backend.common.exception import RelationNotLoadedError
from ai.backend.logging import BraceStyleAdapter
from ai.backend.manager.data.huggingface_registry.types import HuggingFaceRegistryData
from ai.backend.manager.models.base import (
    GUID,
    Base,
)

if TYPE_CHECKING:
    from ai.backend.manager.models.artifact import ArtifactRow
    from ai.backend.manager.models.artifact_registries import ArtifactRegistryRow

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

__all__ = ("HuggingFaceRegistryRow",)


def _get_registry_artifact_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.artifact import ArtifactRow

    return HuggingFaceRegistryRow.id == foreign(ArtifactRow.registry_id)


def _get_registry_meta_join_condition() -> sa.ColumnElement[bool]:
    from ai.backend.manager.models.artifact_registries import ArtifactRegistryRow

    return HuggingFaceRegistryRow.id == foreign(ArtifactRegistryRow.registry_id)


class HuggingFaceRegistryRow(Base):  # type: ignore[misc]
    __tablename__ = "huggingface_registries"

    id: Mapped[uuid.UUID] = mapped_column(
        "id", GUID, primary_key=True, server_default=sa.text("uuid_generate_v4()")
    )
    url: Mapped[str] = mapped_column("url", sa.String, nullable=False)
    token: Mapped[str | None] = mapped_column("token", sa.String, nullable=True, default=None)

    artifacts: Mapped[list[ArtifactRow]] = relationship(
        "ArtifactRow",
        back_populates="huggingface_registry",
        primaryjoin=_get_registry_artifact_join_condition,
        viewonly=True,
    )
    meta: Mapped[ArtifactRegistryRow | None] = relationship(
        "ArtifactRegistryRow",
        back_populates="huggingface_registries",
        primaryjoin=_get_registry_meta_join_condition,
        uselist=False,
        viewonly=True,
    )

    def __str__(self) -> str:
        return f"HuggingFaceRegistryRow(id={self.id}, url={self.url}, token={self.token})"

    def __repr__(self) -> str:
        return self.__str__()

    def to_dataclass(self) -> HuggingFaceRegistryData:
        if self.meta is None:
            raise RelationNotLoadedError()
        return HuggingFaceRegistryData(
            id=self.id, name=self.meta.name, url=self.url, token=self.token
        )



"""Data types for Sokovan scheduler."""

from .allocation import (
    AgentAllocation,
    AllocationBatch,
    KernelAllocation,
    SchedulingFailure,
    SchedulingPredicate,
    SessionAllocation,
)
from .config import NetworkSetup, ScalingGroupInfo, SchedulingConfig
from .image import ImageConfigData, ImageIdentifier
from .lifecycle import (
    KernelBindingData,
    KernelCreationInfo,
    KernelStartData,
    PreparedSessionData,
    PreparedSessionsWithImages,
    ScheduledSessionData,
    ScheduledSessionsWithImages,
    SessionDataForPull,
    SessionDataForStart,
    SessionRunningData,
    SessionsForPullWithImages,
    SessionsForStartWithImages,
    SessionWithKernels,
)
from .result import (
    KernelTerminationInfo,
    PromotionSpec,
    RetryResult,
    SchedulerExecutionError,
    SchedulerExecutionResult,
    SessionStartResult,
    SweepStaleKernelsResult,
)
from .snapshot import (
    AgentOccupancy,
    ConcurrencySnapshot,
    KeypairOccupancy,
    PendingSessionSnapshot,
    ResourceOccupancySnapshot,
    ResourcePolicySnapshot,
    SessionDependencySnapshot,
    SystemSnapshot,
)
from .workload import (
    KernelWorkload,
    KeyPairResourcePolicy,
    PendingSessionInfo,
    SessionDependencyInfo,
    SessionWorkload,
    UserResourcePolicy,
)

__all__ = [
    # allocation
    "AgentAllocation",
    "AllocationBatch",
    "KernelAllocation",
    "SchedulingFailure",
    "SchedulingPredicate",
    "SessionAllocation",
    # config
    "NetworkSetup",
    "ScalingGroupInfo",
    "SchedulingConfig",
    # image
    "ImageConfigData",
    "ImageIdentifier",
    # lifecycle
    "KernelBindingData",
    "KernelCreationInfo",
    "KernelStartData",
    "PreparedSessionData",
    "PreparedSessionsWithImages",
    "ScheduledSessionData",
    "ScheduledSessionsWithImages",
    "SessionDataForPull",
    "SessionDataForStart",
    "SessionRunningData",
    "SessionsForPullWithImages",
    "SessionsForStartWithImages",
    "SessionWithKernels",
    # result
    "KernelTerminationInfo",
    "PromotionSpec",
    "RetryResult",
    "SchedulerExecutionError",
    "SchedulerExecutionResult",
    "SessionStartResult",
    "SweepStaleKernelsResult",
    # snapshot
    "AgentOccupancy",
    "ConcurrencySnapshot",
    "KeypairOccupancy",
    "PendingSessionSnapshot",
    "ResourceOccupancySnapshot",
    "ResourcePolicySnapshot",
    "SessionDependencySnapshot",
    "SystemSnapshot",
    # workload
    "KernelWorkload",
    "KeyPairResourcePolicy",
    "PendingSessionInfo",
    "SessionDependencyInfo",
    "SessionWorkload",
    "UserResourcePolicy",
]



"""Project Fair Share GQL types, filters, and order-by definitions."""

from __future__ import annotations

from datetime import datetime
from decimal import Decimal
from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Any, override
from uuid import UUID

import strawberry
from strawberry import ID, Info
from strawberry.relay import Connection, Edge, Node, NodeID

from ai.backend.manager.api.gql.base import OrderDirection, StringFilter, UUIDFilter
from ai.backend.manager.api.gql.types import GQLFilter, GQLOrderBy, StrawberryGQLContext
from ai.backend.manager.data.fair_share.types import ProjectFairShareData
from ai.backend.manager.data.group.types import ProjectType
from ai.backend.manager.repositories.base import (
    QueryCondition,
    QueryOrder,
    combine_conditions_or,
    negate_conditions,
)
from ai.backend.manager.repositories.fair_share.options import (
    ProjectFairShareConditions,
    ProjectFairShareOrders,
)

from .common import (
    FairShareCalculationSnapshotGQL,
    FairShareSpecGQL,
    ResourceSlotGQL,
)

if TYPE_CHECKING:
    from ai.backend.manager.api.gql.domain_v2.types.node import DomainV2GQL
    from ai.backend.manager.api.gql.project_v2.types.node import ProjectV2GQL
    from ai.backend.manager.api.gql.resource_group.types import ResourceGroupGQL


@strawberry.type(
    name="ProjectFairShare",
    description="Added in 26.1.0. Project-level fair share data representing scheduling priority for a specific project. The fair share factor determines resource allocation relative to other projects in the same domain.",
)
class ProjectFairShareGQL(Node):
    """Project-level fair share data with calculated fair share factor."""

    id: NodeID[str]
    resource_group_name: str = strawberry.field(
        description="Name of the scaling group this fair share belongs to."
    )
    project_id: UUID = strawberry.field(
        description="UUID of the project this fair share is calculated for."
    )
    domain_name: str = strawberry.field(description="Name of the domain the project belongs to.")
    spec: FairShareSpecGQL = strawberry.field(
        description="Fair share specification parameters used for calculation."
    )
    calculation_snapshot: FairShareCalculationSnapshotGQL = strawberry.field(
        description="Snapshot of the most recent fair share calculation results."
    )
    created_at: datetime = strawberry.field(description="Timestamp when this record was created.")
    updated_at: datetime = strawberry.field(
        description="Timestamp when this record was last updated."
    )

    @strawberry.field(  # type: ignore[misc]
        description=("Added in 26.2.0. The project entity associated with this fair share record.")
    )
    async def project(
        self,
        info: Info[StrawberryGQLContext],
    ) -> (
        Annotated[
            ProjectV2GQL,
            strawberry.lazy("ai.backend.manager.api.gql.project_v2.types.node"),
        ]
        | None
    ):
        from ai.backend.manager.api.gql.project_v2.types.node import ProjectV2GQL

        project_data = await info.context.data_loaders.project_loader.load(self.project_id)
        if project_data is None:
            return None
        return ProjectV2GQL.from_data(project_data)

    @strawberry.field(  # type: ignore[misc]
        description=("Added in 26.2.0. The domain entity associated with this fair share record."),
    )
    async def domain(
        self,
        info: Info[StrawberryGQLContext],
    ) -> (
        Annotated[
            DomainV2GQL,
            strawberry.lazy("ai.backend.manager.api.gql.domain_v2.types.node"),
        ]
        | None
    ):
        from ai.backend.manager.api.gql.domain_v2.types.node import DomainV2GQL

        domain_data = await info.context.data_loaders.domain_loader.load(self.domain_name)
        if domain_data is None:
            return None
        return DomainV2GQL.from_data(domain_data)

    @strawberry.field(  # type: ignore[misc]
        description=("Added in 26.2.0. The resource group associated with this fair share record."),
    )
    async def resource_group(
        self,
        info: Info[StrawberryGQLContext],
    ) -> (
        Annotated[
            ResourceGroupGQL,
            strawberry.lazy("ai.backend.manager.api.gql.resource_group.types"),
        ]
        | None
    ):
        from ai.backend.manager.api.gql.resource_group.types import ResourceGroupGQL

        rg_data = await info.context.data_loaders.resource_group_loader.load(
            self.resource_group_name
        )
        if rg_data is None:
            return None
        return ResourceGroupGQL.from_dataclass(rg_data)

    @classmethod
    def from_dataclass(cls, data: ProjectFairShareData) -> ProjectFairShareGQL:
        """Convert ProjectFairShareData to GraphQL type.

        No async needed - Repository provides complete data.
        Note: metadata can be None for default-generated records.
        """
        return cls(
            id=ID(f"{data.resource_group}:{data.project_id}"),
            resource_group_name=data.resource_group,
            project_id=data.project_id,
            domain_name=data.domain_name,
            spec=FairShareSpecGQL.from_spec(
                data.data.spec,
                data.data.use_default,
                data.data.uses_default_resources,
            ),
            calculation_snapshot=FairShareCalculationSnapshotGQL(
                fair_share_factor=data.data.calculation_snapshot.fair_share_factor,
                total_decayed_usage=ResourceSlotGQL.from_resource_slot(
                    data.data.calculation_snapshot.total_decayed_usage
                ),
                normalized_usage=data.data.calculation_snapshot.normalized_usage,
                lookback_start=data.data.calculation_snapshot.lookback_start,
                lookback_end=data.data.calculation_snapshot.lookback_end,
                last_calculated_at=data.data.calculation_snapshot.last_calculated_at,
            ),
            created_at=(
                data.data.metadata.created_at
                if data.data.metadata
                else data.data.calculation_snapshot.last_calculated_at
            ),
            updated_at=(
                data.data.metadata.updated_at
                if data.data.metadata
                else data.data.calculation_snapshot.last_calculated_at
            ),
        )


ProjectFairShareEdge = Edge[ProjectFairShareGQL]


@strawberry.type(
    description=(
        "Added in 26.1.0. Paginated connection for project fair share records. "
        "Provides relay-style cursor-based pagination for efficient traversal of project fair share data. "
        "Use 'edges' to access individual records with cursor information, or 'nodes' for direct data access."
    )
)
class ProjectFairShareConnection(Connection[ProjectFairShareGQL]):
    count: int = strawberry.field(
        description="Total number of project fair share records matching the query criteria."
    )

    def __init__(self, *args: Any, count: int, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.count = count


@strawberry.enum(
    name="ProjectFairShareTypeEnum",
    description="Added in 26.2.0. Project type enum for fair share filtering.",
)
class ProjectFairShareTypeEnum(StrEnum):
    """Project type enum for fair share context."""

    GENERAL = "general"
    MODEL_STORE = "model-store"


@strawberry.input(
    name="ProjectFairShareTypeEnumFilter",
    description=(
        "Added in 26.2.0. Filter for project type enum in fair share queries. "
        "Supports equals, in, not_equals, and not_in operations."
    ),
)
class ProjectFairShareTypeEnumFilter:
    """Filter for project type enum fields in fair share context."""

    equals: ProjectFairShareTypeEnum | None = strawberry.field(
        default=None,
        description="Exact match for project type.",
    )
    in_: list[ProjectFairShareTypeEnum] | None = strawberry.field(
        name="in",
        default=None,
        description="Match any of the provided types.",
    )
    not_equals: ProjectFairShareTypeEnum | None = strawberry.field(
        default=None,
        description="Exclude exact type match.",
    )
    not_in: list[ProjectFairShareTypeEnum] | None = strawberry.field(
        default=None,
        description="Exclude any of the provided types.",
    )


@strawberry.input(
    name="ProjectFairShareProjectNestedFilter",
    description=(
        "Added in 26.2.0. Nested filter for project entity fields in project fair share queries. "
        "Allows filtering by project properties such as name, active status, and type."
    ),
)
class ProjectFairShareProjectNestedFilter:
    """Nested filter for project entity within project fair share."""

    name: StringFilter | None = strawberry.field(
        default=None,
        description="Filter by project name. Supports equals, contains, startsWith, and endsWith.",
    )
    is_active: bool | None = strawberry.field(
        default=None,
        description="Filter by project active status.",
    )
    type: ProjectFairShareTypeEnumFilter | None = strawberry.field(
        default=None,
        description="Filter by project type (GENERAL, MODEL_STORE).",
    )

    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []
        if self.name:
            name_condition = self.name.build_query_condition(
                contains_factory=lambda spec: ProjectFairShareConditions.by_project_name_contains(
                    spec.value
                ),
                equals_factory=lambda spec: ProjectFairShareConditions.by_project_name_equals(
                    spec.value
                ),
                starts_with_factory=lambda spec: ProjectFairShareConditions.by_project_name_starts_with(
                    spec.value
                ),
                ends_with_factory=lambda spec: ProjectFairShareConditions.by_project_name_ends_with(
                    spec.value
                ),
            )
            if name_condition:
                conditions.append(name_condition)
        if self.is_active is not None:
            conditions.append(ProjectFairShareConditions.by_project_is_active(self.is_active))
        if self.type:
            if self.type.equals is not None:
                conditions.append(
                    ProjectFairShareConditions.by_project_type_equals(
                        ProjectType(self.type.equals.value)
                    )
                )
            if self.type.in_ is not None:
                conditions.append(
                    ProjectFairShareConditions.by_project_type_in([
                        ProjectType(t.value) for t in self.type.in_
                    ])
                )
            if self.type.not_equals is not None:
                conditions.append(
                    negate_conditions([
                        ProjectFairShareConditions.by_project_type_equals(
                            ProjectType(self.type.not_equals.value)
                        )
                    ])
                )
            if self.type.not_in is not None:
                conditions.append(
                    negate_conditions([
                        ProjectFairShareConditions.by_project_type_in([
                            ProjectType(t.value) for t in self.type.not_in
                        ])
                    ])
                )
        return conditions


@strawberry.input(
    name="ProjectFairShareFilter",
    description=(
        "Added in 26.1.0. Filter input for querying project fair shares. "
        "Supports filtering by scaling group, project ID, and domain name. "
        "Multiple filters can be combined using AND, OR, and NOT logical operators."
    ),
)
class ProjectFairShareFilter(GQLFilter):
    """Filter for project fair shares."""

    resource_group: StringFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by scaling group name. Scaling groups define resource pool boundaries "
            "where projects compete for resources within their domain. "
            "Supports equals, contains, startsWith, and endsWith operations."
        ),
    )
    project_id: UUIDFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by project UUID. Projects are containers for sessions and user activities. "
            "Supports equals operation for exact match or 'in' operation for multiple UUIDs."
        ),
    )
    domain_name: StringFilter | None = strawberry.field(
        default=None,
        description=(
            "Filter by domain name. This filters projects belonging to a specific domain. "
            "Supports equals, contains, startsWith, and endsWith operations."
        ),
    )
    project: ProjectFairShareProjectNestedFilter | None = strawberry.field(
        default=None,
        description=(
            "Added in 26.2.0. Nested filter for project entity properties. "
            "Allows filtering by project name, active status, and type."
        ),
    )

    AND: list[ProjectFairShareFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with AND logic. All conditions must match.",
    )
    OR: list[ProjectFairShareFilter] | None = strawberry.field(
        default=None,
        description="Combine multiple filters with OR logic. At least one condition must match.",
    )
    NOT: list[ProjectFairShareFilter] | None = strawberry.field(
        default=None,
        description="Negate the specified filters. Records matching these conditions will be excluded.",
    )

    @override
    def build_conditions(self) -> list[QueryCondition]:
        conditions: list[QueryCondition] = []

        if self.resource_group:
            sg_condition = self.resource_group.build_query_condition(
                contains_factory=lambda spec: ProjectFairShareConditions.by_resource_group_contains(
                    spec.value
                ),
                equals_factory=lambda spec: ProjectFairShareConditions.by_resource_group_equals(
                    spec.value
                ),
                starts_with_factory=lambda spec: ProjectFairShareConditions.by_resource_group_starts_with(
                    spec.value
                ),
                ends_with_factory=lambda spec: ProjectFairShareConditions.by_resource_group_ends_with(
                    spec.value
                ),
            )
            if sg_condition:
                conditions.append(sg_condition)

        if self.project_id:
            pid_condition = self.project_id.build_query_condition(
                equals_factory=lambda spec: ProjectFairShareConditions.by_project_id(spec.value),
                in_factory=lambda spec: ProjectFairShareConditions.by_project_ids(spec.values),
            )
            if pid_condition:
                conditions.append(pid_condition)

        if self.domain_name:
            dn_condition = self.domain_name.build_query_condition(
                contains_factory=lambda spec: ProjectFairShareConditions.by_domain_name_contains(
                    spec.value
                ),
                equals_factory=lambda spec: ProjectFairShareConditions.by_domain_name_equals(
                    spec.value
                ),
                starts_with_factory=lambda spec: ProjectFairShareConditions.by_domain_name_starts_with(
                    spec.value
                ),
                ends_with_factory=lambda spec: ProjectFairShareConditions.by_domain_name_ends_with(
                    spec.value
                ),
            )
            if dn_condition:
                conditions.append(dn_condition)

        if self.project:
            conditions.extend(self.project.build_conditions())

        if self.AND:
            for sub_filter in self.AND:
                conditions.extend(sub_filter.build_conditions())

        if self.OR:
            or_conditions: list[QueryCondition] = []
            for sub_filter in self.OR:
                or_conditions.extend(sub_filter.build_conditions())
            if or_conditions:
                conditions.append(combine_conditions_or(or_conditions))

        if self.NOT:
            not_conditions: list[QueryCondition] = []
            for sub_filter in self.NOT:
                not_conditions.extend(sub_filter.build_conditions())
            if not_conditions:
                conditions.append(negate_conditions(not_conditions))

        return conditions


@strawberry.enum(
    name="ProjectFairShareOrderField",
    description=(
        "Added in 26.1.0. Fields available for ordering project fair share query results. "
        "FAIR_SHARE_FACTOR: Order by the calculated fair share factor (0-1 range, lower = higher priority). "
        "CREATED_AT: Order by record creation timestamp. "
        "PROJECT_NAME: Order alphabetically by project name (added in 26.2.0). "
        "PROJECT_IS_ACTIVE: Order by project active status (added in 26.2.0)."
    ),
)
class ProjectFairShareOrderField(StrEnum):
    FAIR_SHARE_FACTOR = "fair_share_factor"
    CREATED_AT = "created_at"
    PROJECT_NAME = "project_name"
    PROJECT_IS_ACTIVE = "project_is_active"


@strawberry.input(
    name="ProjectFairShareOrderBy",
    description=(
        "Added in 26.1.0. Specifies ordering for project fair share query results. "
        "Combine field selection with direction to sort results. "
        "Default direction is DESC (descending)."
    ),
)
class ProjectFairShareOrderBy(GQLOrderBy):
    """OrderBy for project fair shares."""

    field: ProjectFairShareOrderField = strawberry.field(
        description="The field to order by. See ProjectFairShareOrderField for available options."
    )
    direction: OrderDirection = strawberry.field(
        default=OrderDirection.DESC,
        description=(
            "Sort direction. ASC for ascending (lowest first), DESC for descending (highest first). "
            "For fair_share_factor, ASC shows highest priority projects first."
        ),
    )

    @override
    def to_query_order(self) -> QueryOrder:
        ascending = self.direction == OrderDirection.ASC
        match self.field:
            case ProjectFairShareOrderField.FAIR_SHARE_FACTOR:
                return ProjectFairShareOrders.by_fair_share_factor(ascending)
            case ProjectFairShareOrderField.CREATED_AT:
                return ProjectFairShareOrders.by_created_at(ascending)
            case ProjectFairShareOrderField.PROJECT_NAME:
                return ProjectFairShareOrders.by_project_name(ascending)
            case ProjectFairShareOrderField.PROJECT_IS_ACTIVE:
                return ProjectFairShareOrders.by_project_is_active(ascending)


# Mutation Input/Payload Types


@strawberry.input(
    name="UpsertProjectFairShareWeightInput",
    description=(
        "Added in 26.1.0. Input for upserting project fair share weight. "
        "The weight parameter affects scheduling priority - higher weight = higher priority. "
        "Set weight to null to use resource group's default_weight."
    ),
)
class UpsertProjectFairShareWeightInput:
    """Input for upserting project fair share weight."""

    resource_group_name: str = strawberry.field(
        description="Name of the scaling group (resource group) for this fair share."
    )
    project_id: UUID = strawberry.field(description="UUID of the project to update weight for.")
    domain_name: str = strawberry.field(description="Name of the domain the project belongs to.")
    weight: Decimal | None = strawberry.field(
        default=None,
        description=(
            "Priority weight multiplier. Higher weight = higher priority allocation ratio. "
            "Set to null to use resource group's default_weight."
        ),
    )


@strawberry.type(
    name="UpsertProjectFairShareWeightPayload",
    description="Added in 26.1.0. Payload for project fair share weight upsert mutation.",
)
class UpsertProjectFairShareWeightPayload:
    """Payload for project fair share weight upsert mutation."""

    project_fair_share: ProjectFairShareGQL = strawberry.field(
        description="The updated or created project fair share record."
    )


# Bulk Upsert Mutation Input/Payload Types


@strawberry.input(
    name="ProjectWeightInputItem",
    description=(
        "Added in 26.1.0. Input item for a single project weight in bulk upsert. "
        "Represents one project's weight configuration."
    ),
)
class ProjectWeightInputItem:
    """Input item for a single project weight in bulk upsert."""

    project_id: UUID = strawberry.field(description="ID of the project to update weight for.")
    domain_name: str = strawberry.field(description="Name of the domain this project belongs to.")
    weight: Decimal | None = strawberry.field(
        default=None,
        description=(
            "Priority weight multiplier. Higher weight = higher priority allocation ratio. "
            "Set to null to use resource group's default_weight."
        ),
    )


@strawberry.input(
    name="BulkUpsertProjectFairShareWeightInput",
    description=(
        "Added in 26.1.0. Input for bulk upserting project fair share weights. "
        "Allows updating multiple projects in a single transaction."
    ),
)
class BulkUpsertProjectFairShareWeightInput:
    """Input for bulk upserting project fair share weights."""

    resource_group_name: str = strawberry.field(
        description="Name of the scaling group (resource group) for all fair shares."
    )
    inputs: list[ProjectWeightInputItem] = strawberry.field(
        description="List of project weight updates to apply."
    )


@strawberry.type(
    name="BulkUpsertProjectFairShareWeightPayload",
    description="Added in 26.1.0. Payload for bulk project fair share weight upsert mutation.",
)
class BulkUpsertProjectFairShareWeightPayload:
    """Payload for bulk project fair share weight upsert mutation."""

    upserted_count: int = strawberry.field(
        description="Number of project fair share records created or updated."
    )



"""add relation_type to association_scopes_entities

Revision ID: 8fd6f47bd226
Revises: 321c033a321b
Create Date: 2026-02-10 19:44:48.369502

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8fd6f47bd226"
down_revision = "321c033a321b"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "association_scopes_entities",
        sa.Column("relation_type", sa.String(32), nullable=False, server_default="auto"),
    )


def downgrade() -> None:
    op.drop_column("association_scopes_entities", "relation_type")



import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
from typing import TypeVar

from ai.backend.common.data.permission.types import EntityType
from ai.backend.common.exception import ErrorCode
from ai.backend.manager.actions.types import ActionOperationType, ActionSpec, OperationStatus
from ai.backend.manager.data.common.types import SearchResult


class BaseAction(ABC):
    def entity_id(self) -> str | None:
        """
        Return the ID of the entity this action operates on.
        This returns `None` by default because subclasses may not always need to specify an entity ID.
        """
        return None

    @classmethod
    @abstractmethod
    def entity_type(cls) -> EntityType:
        raise NotImplementedError

    @classmethod
    @abstractmethod
    def operation_type(cls) -> ActionOperationType:
        raise NotImplementedError

    @classmethod
    def spec(cls) -> ActionSpec:
        return ActionSpec(
            entity_type=cls.entity_type(),
            operation_type=cls.operation_type(),
        )


@dataclass
class BaseActionTriggerMeta:
    action_id: uuid.UUID
    started_at: datetime


class BaseActionResult(ABC):
    @abstractmethod
    def entity_id(self) -> str | None:
        raise NotImplementedError


@dataclass
class SearchActionResult[T](BaseActionResult):
    result: SearchResult[T]

    def entity_id(self) -> str | None:
        return None


@dataclass
class BaseActionResultMeta:
    action_id: uuid.UUID
    entity_id: str | None
    status: OperationStatus
    description: str
    started_at: datetime
    ended_at: datetime
    duration: timedelta
    error_code: ErrorCode | None


TAction = TypeVar("TAction", bound=BaseAction)
TActionResult = TypeVar("TActionResult", bound=BaseActionResult)


@dataclass
class ProcessResult:
    meta: BaseActionResultMeta



"""CreatorSpec implementations for group repository."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.common.types import ResourceSlot, VFolderHostPermissionMap
from ai.backend.manager.models.group import GroupRow, ProjectType
from ai.backend.manager.repositories.base import CreatorSpec


@dataclass
class GroupCreatorSpec(CreatorSpec[GroupRow]):
    """CreatorSpec for group creation."""

    name: str
    domain_name: str
    type: ProjectType | None = None
    description: str | None = None
    is_active: bool | None = None
    total_resource_slots: ResourceSlot | None = None
    allowed_vfolder_hosts: VFolderHostPermissionMap | None = None
    integration_id: str | None = None
    resource_policy: str | None = None
    container_registry: dict[str, str] | None = None
    dotfiles: bytes | None = None

    @override
    def build_row(self) -> GroupRow:
        return GroupRow(
            name=self.name,
            domain_name=self.domain_name,
            type=self.type or ProjectType.GENERAL,
            description=self.description,
            is_active=self.is_active if self.is_active is not None else True,
            total_resource_slots=self.total_resource_slots or ResourceSlot(),
            allowed_vfolder_hosts=self.allowed_vfolder_hosts or VFolderHostPermissionMap(),
            integration_id=self.integration_id,
            resource_policy=self.resource_policy,
            dotfiles=self.dotfiles,
            container_registry=self.container_registry,
        )



from __future__ import annotations

from dataclasses import dataclass, field
from enum import Enum
from uuid import UUID

from ai.backend.common.data.model_deployment.types import (
    ActivenessStatus,
    LivenessStatus,
    ModelDeploymentStatus,
    ReadinessStatus,
)
from ai.backend.manager.api.gql.base import StringFilter
from ai.backend.manager.data.deployment.types import (
    AccessTokenOrderField,
    DeploymentOrderField,
    ModelRevisionOrderField,
    ReplicaOrderField,
)


@dataclass
class DeploymentOrderingOptions:
    """Ordering options for deployment queries."""

    order_by: list[tuple[DeploymentOrderField, bool]] = field(
        default_factory=lambda: [(DeploymentOrderField.CREATED_AT, True)]
    )  # (field, desc)


@dataclass
class ModelRevisionOrderingOptions:
    """Ordering options for model revision queries."""

    order_by: list[tuple[ModelRevisionOrderField, bool]] = field(
        default_factory=lambda: [(ModelRevisionOrderField.CREATED_AT, True)]
    )  # (field, desc)


@dataclass
class ModelReplicaOrderingOptions:
    """Ordering options for model replica queries."""

    order_by: list[tuple[ReplicaOrderField, bool]] = field(
        default_factory=lambda: [(ReplicaOrderField.CREATED_AT, True)]
    )  # (field, desc)


@dataclass
class AccessTokenOrderingOptions:
    """Ordering options for access token queries."""

    order_by: list[tuple[AccessTokenOrderField, bool]] = field(
        default_factory=lambda: [(AccessTokenOrderField.CREATED_AT, True)]
    )  # (field, desc)


class DeploymentStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class DeploymentStatusFilter:
    """Status filter with operation type and values."""

    type: DeploymentStatusFilterType
    values: list[ModelDeploymentStatus]


@dataclass
class DeploymentFilterOptions:
    """Filtering options for deployments."""

    name: StringFilter | None = None
    status: DeploymentStatusFilter | None = None
    open_to_public: bool | None = None
    tags: StringFilter | None = None
    endpoint_url: StringFilter | None = None
    id: UUID | None = None

    # Logical operations
    AND: list[DeploymentFilterOptions] | None = None
    OR: list[DeploymentFilterOptions] | None = None
    NOT: list[DeploymentFilterOptions] | None = None


@dataclass
class ModelRevisionFilterOptions:
    """Filtering options for model revisions."""

    name: StringFilter | None = None
    deployment_id: UUID | None = None
    id: UUID | None = None
    ids_in: list[UUID] | None = None

    # Logical operations
    AND: list[ModelRevisionFilterOptions] | None = None
    OR: list[ModelRevisionFilterOptions] | None = None
    NOT: list[ModelRevisionFilterOptions] | None = None


class ReadinessStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class ReadinessStatusFilter:
    """Readiness status filter with operation type and values."""

    type: ReadinessStatusFilterType
    values: list[ReadinessStatus]


class LivenessStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class LivenessStatusFilter:
    """Liveness status filter with operation type and values."""

    type: LivenessStatusFilterType
    values: list[LivenessStatus]


class ActivenessStatusFilterType(Enum):
    IN = "in"
    EQUALS = "equals"


@dataclass
class ActivenessStatusFilter:
    """Activeness status filter with operation type and values."""

    type: ActivenessStatusFilterType
    values: list[ActivenessStatus]


@dataclass
class ModelReplicaFilterOptions:
    """Filtering options for model replicas."""

    readiness_status_filter: ReadinessStatusFilter | None = None
    liveness_status_filter: LivenessStatusFilter | None = None
    activeness_status_filter: ActivenessStatusFilter | None = None
    id: UUID | None = None
    ids_in: list[UUID] | None = None

    # Logical operations
    AND: list[ModelReplicaFilterOptions] | None = None
    OR: list[ModelReplicaFilterOptions] | None = None
    NOT: list[ModelReplicaFilterOptions] | None = None



"""fix duplicate enum values

Revision ID: 38741c85f228
Revises: 31463788c713
Create Date: 2024-05-27 15:33:27.394135

"""

import sqlalchemy as sa
from alembic import op
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = "38741c85f228"
down_revision = "31463788c713"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    conn.exec_driver_sql("ALTER TYPE routestatus ADD VALUE 'terminating'")
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    conn = op.get_bind()
    conn.exec_driver_sql(
        "UPDATE routings SET status = 'unhealthy'::routestatus WHERE status = 'terminating'::routestatus"
    )
    op.alter_column("routings", "status", type_=sa.TEXT)

    routestatus_choices = ["healthy", "unhealthy", "failed_to_start", "provisioning"]
    routestatus = postgresql.ENUM(*routestatus_choices, name="routestatus")
    routestatus.drop(conn)
    routestatus.create(conn)

    op.alter_column("routings", "status", type_=routestatus, postgresql_using="status::routestatus")

    # ### end Alembic commands ###



"""rename_kernel_dependencies_to_session_dependencies

Revision ID: e421c02cf9e4
Revises: 548cc8aa49c8
Create Date: 2020-09-14 10:45:40.218548

"""

from alembic import op
from sqlalchemy.sql import text

# revision identifiers, used by Alembic.
revision = "e421c02cf9e4"
down_revision = "548cc8aa49c8"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.drop_constraint("fk_kernel_dependencies_depends_on_kernels", "kernel_dependencies")
    op.drop_constraint("fk_kernel_dependencies_kernel_id_kernels", "kernel_dependencies")
    op.rename_table("kernel_dependencies", "session_dependencies")
    op.alter_column("session_dependencies", "kernel_id", new_column_name="session_id")
    op.execute(text("ALTER INDEX pk_kernel_dependencies RENAME TO pk_session_dependencies"))
    op.execute(
        text(
            "ALTER INDEX ix_kernel_dependencies_depends_on "
            "RENAME TO ix_session_dependencies_depends_on"
        )
    )
    op.execute(
        text(
            "ALTER INDEX ix_kernel_dependencies_kernel_id "
            "RENAME TO ix_session_dependencies_session_id"
        )
    )
    # NOTE: we keep the fkey target as "kernels.id" instead of "kernels.session_id"
    #       because fkey target must be a unique index and in Backend.AI `kernels.session_id`
    #       is same to the main kernel's `kernels.id`.
    op.create_foreign_key(
        None,
        "session_dependencies",
        "kernels",
        ["session_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        None,
        "session_dependencies",
        "kernels",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )


def downgrade() -> None:
    op.drop_constraint("fk_session_dependencies_depends_on_kernels", "session_dependencies")
    op.drop_constraint("fk_session_dependencies_session_id_kernels", "session_dependencies")
    op.rename_table("session_dependencies", "kernel_dependencies")
    op.alter_column("kernel_dependencies", "session_id", new_column_name="kernel_id")
    op.execute(text("ALTER INDEX pk_session_dependencies RENAME TO pk_kernel_dependencies"))
    op.execute(
        text(
            "ALTER INDEX ix_session_dependencies_depends_on "
            "RENAME TO ix_kernel_dependencies_depends_on"
        )
    )
    op.execute(
        text(
            "ALTER INDEX ix_session_dependencies_session_id "
            "RENAME TO ix_kernel_dependencies_kernel_id"
        )
    )
    op.create_foreign_key(
        None,
        "kernel_dependencies",
        "kernels",
        ["kernel_id"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )
    op.create_foreign_key(
        None,
        "kernel_dependencies",
        "kernels",
        ["depends_on"],
        ["id"],
        onupdate="CASCADE",
        ondelete="CASCADE",
    )



from __future__ import annotations

import enum
from dataclasses import dataclass, field
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, NamedTuple
from uuid import UUID

from ai.backend.common.types import CIStrEnum, ImageCanonical, ImageID

if TYPE_CHECKING:
    from ai.backend.manager.models.image import Resources


class ImageStatus(enum.StrEnum):
    ALIVE = "ALIVE"
    DELETED = "DELETED"


class ImageOrderField(enum.StrEnum):
    NAME = "NAME"
    CREATED_AT = "CREATED_AT"


class ImageType(CIStrEnum):
    COMPUTE = "compute"
    SYSTEM = "system"
    SERVICE = "service"


class ImageIdentifier(NamedTuple):
    """
    Represent a tuple of image's canonical string and architecture, uniquely corresponding to an ImageRow.
    """

    canonical: str
    architecture: str


@dataclass
class ImageLabelsData:
    label_data: dict[str, str]


@dataclass
class ImageResourcesData:
    resources_data: Resources


@dataclass
class ImageTagEntry:
    """A single parsed tag component from the image reference."""

    key: str
    value: str


@dataclass
class ResourceLimit:
    key: str
    min: Decimal
    max: Decimal


@dataclass
class ImageData:
    id: ImageID = field(compare=False)
    name: ImageCanonical
    project: str | None
    image: str
    created_at: datetime | None = field(compare=False)
    tag: str | None
    registry: str
    registry_id: UUID
    architecture: str
    config_digest: str
    size_bytes: int
    is_local: bool
    type: ImageType
    accelerators: str | None
    labels: ImageLabelsData
    resources: ImageResourcesData
    resource_limits: list[ResourceLimit]
    tags: list[ImageTagEntry]
    status: ImageStatus


@dataclass
class KVPair:
    key: str
    value: str


@dataclass
class ResourceLimitInput:
    """Input for setting a resource limit with optional min/max values."""

    slot_name: str
    min_value: Decimal | None
    max_value: Decimal | None


@dataclass
class ImageDataWithDetails:
    id: ImageID = field(compare=False)
    name: ImageCanonical
    namespace: str
    base_image_name: str
    project: str
    humanized_name: str
    tag: str | None
    tags: list[KVPair]
    version: str | None
    registry: str
    registry_id: UUID
    type: ImageType
    architecture: str
    is_local: bool
    status: ImageStatus
    resource_limits: list[ResourceLimit]
    supported_accelerators: list[str] = field(default_factory=list)
    digest: str | None = field(default=None)
    labels: list[KVPair] = field(default_factory=list)
    aliases: list[str] = field(default_factory=list)
    size_bytes: int = field(default=0)
    created_at: datetime | None = field(default=None)
    # legacy
    hash: str | None = field(default=None)


@dataclass
class ImageAgentInstallStatus:
    """
    Represents the installation status of an image on agents.
    """

    installed: bool
    agent_names: list[str] = field(default_factory=list)


@dataclass
class ImageWithAgentInstallStatus:
    """
    Wraps detailed image information and its agent installation status.
    """

    image: ImageDataWithDetails
    agent_install_status: ImageAgentInstallStatus


@dataclass
class RescanImagesResult:
    images: list[ImageData]
    errors: list[str] = field(default_factory=list)


@dataclass
class ImageAliasData:
    id: UUID = field(compare=False)
    alias: str


@dataclass
class ImageListResult:
    """Search result with total count and pagination info for images."""

    items: list[ImageData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool


@dataclass
class ImageAliasListResult:
    """Search result with total count and pagination info for image aliases."""

    items: list[ImageAliasData]
    image_ids: list[ImageID]
    total_count: int
    has_next_page: bool
    has_previous_page: bool



from __future__ import annotations

import logging
from collections.abc import Mapping

from ai.backend.common.clients.valkey_client.valkey_image.client import ValkeyImageClient
from ai.backend.common.types import AgentId, ImageID
from ai.backend.logging.utils import BraceStyleAdapter

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class ImageStatefulSource:
    """
    Source for stateful image-related operations.
    Uses simple JSON serialization with short TTL for caching.
    """

    _valkey_image: ValkeyImageClient

    def __init__(self, valkey_image: ValkeyImageClient) -> None:
        self._valkey_image = valkey_image

    async def list_agents_with_image(self, image_id: ImageID) -> set[AgentId]:
        return await self._valkey_image.get_agents_for_image(image_id)

    async def list_agents_with_images(
        self, image_ids: list[ImageID]
    ) -> Mapping[ImageID, set[AgentId]]:
        return await self._valkey_image.get_agents_for_images(image_ids)



from typing import override

from ai.backend.manager.actions.monitors.monitor import ActionMonitor
from ai.backend.manager.actions.processor.scope import ScopeActionProcessor
from ai.backend.manager.actions.processor.single_entity import SingleEntityActionProcessor
from ai.backend.manager.actions.types import AbstractProcessorPackage, ActionSpec
from ai.backend.manager.services.vfolder.actions.base import (
    CloneVFolderAction,
    CloneVFolderActionResult,
    CreateVFolderAction,
    CreateVFolderActionResult,
    DeleteForeverVFolderAction,
    DeleteForeverVFolderActionResult,
    ForceDeleteVFolderAction,
    ForceDeleteVFolderActionResult,
    GetTaskLogsAction,
    GetTaskLogsActionResult,
    GetVFolderAction,
    GetVFolderActionResult,
    ListVFolderAction,
    ListVFolderActionResult,
    MoveToTrashVFolderAction,
    MoveToTrashVFolderActionResult,
    PurgeVFolderAction,
    PurgeVFolderActionResult,
    RestoreVFolderFromTrashAction,
    RestoreVFolderFromTrashActionResult,
    UpdateVFolderAttributeAction,
    UpdateVFolderAttributeActionResult,
)
from ai.backend.manager.services.vfolder.services.vfolder import VFolderService


class VFolderProcessors(AbstractProcessorPackage):
    create_vfolder: ScopeActionProcessor[CreateVFolderAction, CreateVFolderActionResult]
    get_vfolder: SingleEntityActionProcessor[GetVFolderAction, GetVFolderActionResult]
    list_vfolder: ScopeActionProcessor[ListVFolderAction, ListVFolderActionResult]
    update_vfolder_attribute: SingleEntityActionProcessor[
        UpdateVFolderAttributeAction, UpdateVFolderAttributeActionResult
    ]
    move_to_trash_vfolder: SingleEntityActionProcessor[
        MoveToTrashVFolderAction, MoveToTrashVFolderActionResult
    ]
    restore_vfolder_from_trash: SingleEntityActionProcessor[
        RestoreVFolderFromTrashAction, RestoreVFolderFromTrashActionResult
    ]
    delete_forever_vfolder: SingleEntityActionProcessor[
        DeleteForeverVFolderAction, DeleteForeverVFolderActionResult
    ]
    purge_vfolder: SingleEntityActionProcessor[PurgeVFolderAction, PurgeVFolderActionResult]
    force_delete_vfolder: SingleEntityActionProcessor[
        ForceDeleteVFolderAction, ForceDeleteVFolderActionResult
    ]
    clone_vfolder: SingleEntityActionProcessor[CloneVFolderAction, CloneVFolderActionResult]
    get_task_logs: SingleEntityActionProcessor[GetTaskLogsAction, GetTaskLogsActionResult]

    def __init__(self, service: VFolderService, action_monitors: list[ActionMonitor]) -> None:
        self.create_vfolder = ScopeActionProcessor(service.create, action_monitors)
        self.get_vfolder = SingleEntityActionProcessor(service.get, action_monitors)
        self.list_vfolder = ScopeActionProcessor(service.list, action_monitors)
        self.update_vfolder_attribute = SingleEntityActionProcessor(
            service.update_attribute, action_monitors
        )
        self.move_to_trash_vfolder = SingleEntityActionProcessor(
            service.move_to_trash, action_monitors
        )
        self.restore_vfolder_from_trash = SingleEntityActionProcessor(
            service.restore, action_monitors
        )
        self.delete_forever_vfolder = SingleEntityActionProcessor(
            service.delete_forever, action_monitors
        )
        self.purge_vfolder = SingleEntityActionProcessor(service.purge, action_monitors)
        self.force_delete_vfolder = SingleEntityActionProcessor(
            service.force_delete, action_monitors
        )
        self.clone_vfolder = SingleEntityActionProcessor(service.clone, action_monitors)
        self.get_task_logs = SingleEntityActionProcessor(service.get_task_logs, action_monitors)

    @override
    def supported_actions(self) -> list[ActionSpec]:
        return [
            CreateVFolderAction.spec(),
            GetVFolderAction.spec(),
            ListVFolderAction.spec(),
            UpdateVFolderAttributeAction.spec(),
            MoveToTrashVFolderAction.spec(),
            RestoreVFolderFromTrashAction.spec(),
            DeleteForeverVFolderAction.spec(),
            PurgeVFolderAction.spec(),
            ForceDeleteVFolderAction.spec(),
            CloneVFolderAction.spec(),
            GetTaskLogsAction.spec(),
        ]



"""add runtime_variant column

Revision ID: 5d92c9cc930c
Revises: 679e5721e94d
Create Date: 2024-06-05 15:47:57.541242

"""

import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "5d92c9cc930c"
down_revision = "679e5721e94d"
branch_labels = None
depends_on = None


def upgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "endpoints",
        sa.Column("runtime_variant", sa.VARCHAR, nullable=True, server_default="custom"),
    )
    conn = op.get_bind()
    conn.exec_driver_sql("UPDATE endpoints SET runtime_variant = 'custom';")
    op.alter_column("endpoints", "runtime_variant", nullable=False)
    # ### end Alembic commands ###


def downgrade() -> None:
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column("endpoints", "runtime_variant")
    # ### end Alembic commands ###



"""Search domains action and result types."""

from __future__ import annotations

from dataclasses import dataclass
from typing import override

from ai.backend.manager.actions.action import BaseActionResult
from ai.backend.manager.actions.types import ActionOperationType
from ai.backend.manager.data.domain.types import DomainData
from ai.backend.manager.repositories.base import BatchQuerier
from ai.backend.manager.services.domain.actions.base import DomainAction


@dataclass
class SearchDomainsAction(DomainAction):
    """Action to search domains with filtering, ordering, and pagination.

    Args:
        querier: BatchQuerier containing filters, orders, and pagination options.
    """

    querier: BatchQuerier

    @override
    def entity_id(self) -> str | None:
        return None

    @override
    @classmethod
    def operation_type(cls) -> ActionOperationType:
        return ActionOperationType.SEARCH


@dataclass
class SearchDomainsActionResult(BaseActionResult):
    """Result of SearchDomainsAction.

    Args:
        items: List of domain data items in the current page.
        total_count: Total number of domains matching the filter.
        has_next_page: Whether there are more items after the current page.
        has_previous_page: Whether there are items before the current page.
    """

    items: list[DomainData]
    total_count: int
    has_next_page: bool
    has_previous_page: bool

    @override
    def entity_id(self) -> str | None:
        return None



"""User GraphQL enum types."""

from __future__ import annotations

from enum import StrEnum

import strawberry


@strawberry.enum(
    name="UserStatusV2",
    description=(
        "Added in 26.2.0. User account status. "
        "ACTIVE: User can log in and use the system. "
        "INACTIVE: User account is disabled but preserved. "
        "DELETED: User has been soft-deleted. "
        "BEFORE_VERIFICATION: User account is pending email verification."
    ),
)
class UserStatusEnumGQL(StrEnum):
    ACTIVE = "active"
    INACTIVE = "inactive"
    DELETED = "deleted"
    BEFORE_VERIFICATION = "before-verification"


@strawberry.enum(
    name="UserRoleV2",
    description=(
        "Added in 26.2.0. User role determining access permissions. "
        "USER: Standard user with basic permissions. "
        "ADMIN: Domain administrator with elevated permissions. "
        "SUPERADMIN: System-wide administrator with full access. "
        "MONITOR: Read-only access for monitoring purposes."
    ),
)
class UserRoleEnumGQL(StrEnum):
    USER = "user"
    ADMIN = "admin"
    SUPERADMIN = "superadmin"
    MONITOR = "monitor"
